import os
import random
import secrets
from functools import partial
from typing import Callable, Literal, Optional, Union

import torch
from torch import distributed as dist

from lhotse.utils import fix_random_seed

LHOTSE_PROCESS_SEED = "LHOTSE_PROCESS_SEED"


def make_worker_init_fn(
    rank: Optional[int] = None,
    world_size: Optional[int] = None,
    set_different_node_and_worker_seeds: bool = True,
    seed: Optional[int] = 42,
) -> Optional[Callable[[int], None]]:
    """
    Calling this function creates a worker_init_fn suitable to pass to PyTorch's DataLoader.

    It helps with two issues:

    * sets the random seeds differently for each worker and node, which helps with
        avoiding duplication in randomized data augmentation techniques.
    * sets environment variables that help WebDataset detect it's inside multi-GPU (DDP)
        training, so that it correctly de-duplicates the data across nodes.
    """
    return partial(
        worker_init_fn,
        rank=rank,
        world_size=world_size,
        set_different_node_and_worker_seeds=set_different_node_and_worker_seeds,
        seed=seed,
    )


def worker_init_fn(
    worker_id: int,
    rank: Optional[int] = None,
    world_size: Optional[int] = None,
    set_different_node_and_worker_seeds: bool = True,
    seed: Optional[int] = 42,
) -> None:
    """
    Function created by :func:`~lhotse.dataset.dataloading.make_worker_init_fn`, refer to its documentation for details.
    """
    if set_different_node_and_worker_seeds:
        process_seed = seed + 100 * worker_id
        if rank is not None:
            process_seed += 100000 * rank
        fix_random_seed(process_seed)
        os.environ[LHOTSE_PROCESS_SEED] = str(process_seed)

    if rank is None and world_size is None:
        return

    assert (
        rank is not None and world_size is not None
    ), f"Both args must be not None: rank={rank}, world_size={world_size}"

    # This sets the rank/world_size info for WebDataset to read it in worker subprocesses.
    # If we didn't do it, WebDataset will "think" this is always single-node training,
    # because DataLoader workers did not initialize torch.distributed.
    os.environ["RANK"] = str(rank)
    os.environ["WORLD_SIZE"] = str(world_size)


def resolve_seed(seed: Union[int, Literal["trng", "randomized"]]) -> int:
    """
    Resolves the special values of random seed supported in Lhotse.

    If it's an integer, we'll just return it.

    If it's "trng", we'll use the ``secrets`` module to generate a random seed
    using a true RNG (to the extend supported by the OS).

    If it's "randomized", we'll check whether we're in a dataloading worker of ``torch.utils.data.DataLoader``.
    If we are, we expect that it was passed the result of :func:`~lhotse.dataset.dataloading.make_worker_init_fn`
    into its ``worker_init_fn`` argument, in which case we'll return a special seed exclusive to that worker.
    If we are not in a dataloading worker (or ``num_workers`` was set to ``0``), we'll return Python's ``random``
    module global seed.
    """
    if isinstance(seed, int):
        return seed

    if seed == "randomized":
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            # not in a dataloader sub-process: get python global random seed
            return random.getstate()[1][0]
        else:
            # in a dataloader sub-process: read out the seed we assigned to it
            assert LHOTSE_PROCESS_SEED in os.environ, (
                "Requested seed='randomized' for shuffling shards differently "
                "on each DataLoader node and worker, "
                "but lhotse.dataset.dataloading.worker_init_fn was not called."
            )
            return int(os.environ[LHOTSE_PROCESS_SEED])

    if seed == "trng":
        return secrets.randbelow(2**32)

    raise ValueError(
        f"Unexpected type or value of seed: {type(seed)=} {seed=}. "
        f"Supported values are: int, 'trng', and 'randomized'."
    )


def get_world_size() -> int:
    """Source: https://github.com/danpovey/icefall/blob/74bf02bba6016c1eb37858a4e0e8a40f7d302bdb/icefall/dist.py#L56"""
    if "WORLD_SIZE" in os.environ:
        return int(os.environ["WORLD_SIZE"])
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size()
    else:
        return 1


def get_rank() -> int:
    """Source: https://github.com/danpovey/icefall/blob/74bf02bba6016c1eb37858a4e0e8a40f7d302bdb/icefall/dist.py#L56"""
    if "RANK" in os.environ:
        return int(os.environ["RANK"])
    elif dist.is_available() and dist.is_initialized():
        return dist.get_rank()
    else:
        return 0



import logging

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.audio_mnist import download_audio_mnist, prepare_audio_mnist
from lhotse.utils import Pathlike


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "--force-download",
    type=bool,
    default=False,
    help="If True, download even if file is present.",
)
def audio_mnist(target_dir: Pathlike, force_download: bool):
    """AudioMNIST dataset download."""
    download_audio_mnist(target_dir, force_download)


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def audio_mnist(corpus_dir: Pathlike, output_dir: Pathlike):
    """AudioMNIST corpus data preparation."""
    logging.basicConfig(level=logging.INFO)
    prepare_audio_mnist(
        corpus_dir,
        output_dir=output_dir,
    )



import logging
import warnings
from dataclasses import dataclass
from functools import partial, reduce
from itertools import groupby
from operator import add
from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch

from lhotse.audio import Recording
from lhotse.cut.data import DataCut
from lhotse.features import Features
from lhotse.supervision import SupervisionSegment
from lhotse.utils import (
    add_durations,
    fastcopy,
    hash_str_to_int,
    is_equal_or_contains,
    merge_items_with_delimiter,
    overlaps,
    rich_exception_info,
    to_list,
    uuid4,
)


@dataclass
class MultiCut(DataCut):
    """
    :class:`~lhotse.cut.MultiCut` is a :class:`~lhotse.cut.Cut` that is analogous to the MonoCut.
    While MonoCut represents a single channel of a recording, MultiCut represents multi-channel
    recordings where supervisions may or may not be shared across channels.
    It is intended to be used to store, for example, segments of a microphone array recording.
    The following diagrams illustrate some examples for MultiCut usage:

    >>> 2-channel telephone recording with 2 supervisions, one for each channel (e.g., Switchboard):


                  ╔══════════════════════════════  MultiCut  ═════════════════╗
                  ║ ┌──────────────────────────┐                              ║
     Channel 1  ──╬─│   Hello this is John.    │──────────────────────────────╬────────
                  ║ └──────────────────────────┘                              ║
                  ║                               ┌──────────────────────────┐║
     Channel 2  ──╬───────────────────────────────│ Hey, John. How are you?  │╠────────
                  ║                               └──────────────────────────┘║
                  ╚═══════════════════════════════════════════════════════════╝

    >>> Multi-array multi-microphone recording with shared supervisions (e.g., CHiME-6),
    along with close-talk microphones (A and B are distant arrays, C is close-talk):

               ╔═══════════════════════════════════════════════════════════════════════════╗
               ║ ┌───────────────────┐                         ┌───────────────────┐       ║
       A-1   ──╬─┤                   ├─────────────────────────┤                   ├───────╬─
               ║ │ What did you do?  │                         │I cleaned my room. │       ║
       A-2   ──╬─┤                   ├─────────────────────────┤                   ├───────╬─
               ║ └───────────────────┘  ┌───────────────────┐  └───────────────────┘       ║
       B-1   ──╬────────────────────────┤Yeah, we were going├──────────────────────────────╬─
               ║                        │   to the mall.    │                              ║
       B-2   ──╬────────────────────────┤                   ├──────────────────────────────╬─
               ║                        └───────────────────┘        ┌───────────────────┐ ║
        C    ──╬─────────────────────────────────────────────────────┤      Right.       ├─╬─
               ║                                                     └───────────────────┘ ║
               ╚════════════════════════════════  MultiCut  ═══════════════════════════════╝

    By definition, a MultiCut has the same attributes as a MonoCut. The key difference is that
    the Recording object has multiple channels, and the Supervision objects may correspond to
    any of these channels. The channels that the MultiCut can be a subset of the Recording
    channels, but must be a superset of the Supervision channels.

    See also:

        - :class:`lhotse.cut.Cut`
        - :class:`lhotse.cut.MonoCut`
        - :class:`lhotse.cut.CutSet`
        - :class:`lhotse.cut.MixedCut`
    """

    channel: List[int]

    @property
    def num_channels(self) -> int:
        return len(to_list(self.channel))

    @rich_exception_info
    def load_features(
        self, channel: Optional[Union[int, List[int]]] = None
    ) -> Optional[np.ndarray]:
        """
        Load the features from the underlying storage and cut them to the relevant
        [begin, duration] region of the current MultiCut.

        :param channel: The channel to load the features for. If None, all channels will be loaded.
            This is useful for the case when we have features extracted for each channel of
            the multi-cut, and we want to selectively load them.
        """
        if self.has_features:
            feats = self.features.load(
                start=self.start,
                duration=self.duration,
                channel_id=self.channel if channel is None else channel,
            )
            # Note: we forgive off-by-one errors in the feature matrix frames
            #       due to various hard-to-predict floating point arithmetic issues.
            #       If needed, we will remove or duplicate the last frame to be
            #       consistent with the manifests declared "num_frames".
            if feats.shape[0] - self.num_frames == 1:
                feats = feats[: self.num_frames, ...]
            elif feats.shape[0] - self.num_frames == -1:
                feats = np.concatenate((feats, feats[-1:, ...]), axis=0)
            return feats
        return None

    @rich_exception_info
    def load_audio(
        self, channel: Optional[Union[int, List[int]]] = None
    ) -> Optional[np.ndarray]:
        """
        Load the audio by locating the appropriate recording in the supplied Recording.
        The audio is trimmed to the [begin, end] range specified by the MultiCut.

        :param channel: optional int or list of int, the subset of channels to load (all by default).
        :return: a numpy ndarray with audio samples, with shape (C <channel>, N <samples>)
        """
        if self.has_recording:
            return self.recording.load_audio(
                channels=self.channel if channel is None else channel,
                offset=self.start,
                duration=self.duration,
            )
        return None

    @rich_exception_info
    def load_video(
        self,
        channel: Optional[Union[int, List[int]]] = None,
        with_audio: bool = True,
    ) -> Optional[Tuple[torch.Tensor, Optional[torch.Tensor]]]:
        """
        Load the subset of video (and audio) from attached recording.
        The data is trimmed to the [begin, end] range specified by the MonoCut.

        :param channel: optional int or list of int, the subset of channels to load (all by default).
        :param with_audio: bool, whether to load and return audio alongside video. True by default.
        :return: a tuple of video tensor and optionally audio tensor (or ``None``),
            or ``None`` if this cut has no video.
        """
        if self.has_video:
            return self.recording.load_video(
                channels=self.channel if channel is None else channel,
                offset=self.start,
                duration=self.duration,
                with_audio=with_audio,
            )
        return None

    def reverb_rir(
        self,
        rir_recording: Optional[Union[Recording, DataCut]] = None,
        normalize_output: bool = True,
        early_only: bool = False,
        affix_id: bool = True,
        rir_channels: Sequence[int] = (0,),
        room_rng_seed: Optional[int] = None,
        source_rng_seed: Optional[int] = None,
    ) -> "MultiCut":
        """
        Return a new ``MultiCut`` that will convolve the audio with the provided impulse response.
        If the `rir_recording` is multi-channel, the `rir_channels` argument determines which channels
        will be used. This list must be of the same length as the number of channels in the `MultiCut`.

        If no ``rir_recording`` is provided, we will generate an impulse response using a fast random
        generator (https://arxiv.org/abs/2208.04101), only if the MultiCut has exactly one channel.
        At the moment we do not support simulation of multi-channel impulse responses.

        :param rir_recording: The impulse response to use for convolving.
        :param normalize_output: When true, output will be normalized to have energy as input.
        :param early_only: When true, only the early reflections (first 50 ms) will be used.
        :param affix_id: When true, we will modify the ``MonoCut.id`` field
            by affixing it with "_rvb".
        :param rir_channels: The channels of the impulse response to use. First channel is used by default.
            If multiple channels are specified, this will produce a MixedCut instead of a MonoCut.
        :param room_rng_seed: The seed for the room configuration.
        :param source_rng_seed: The seed for the source positions.
        :return: a modified copy of the current ``MonoCut``.
        """
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot apply reverberation on a MultiCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to reverberate a MultiCut that references pre-computed features. "
                "The feature manifest will be detached, as we do not support feature-domain "
                "reverberation."
            )
            self.features = None

        if rir_recording is None:
            assert self.num_channels == 1, (
                "We do not support reverberation simulation for multi-channel recordings. "
                "Please provide an impulse response."
            )
            if room_rng_seed is None:
                room_rng_seed = hash_str_to_int(str(uuid4()) + self.id)
            if source_rng_seed is None:
                source_rng_seed = room_rng_seed
        else:
            assert all(
                c < rir_recording.num_channels for c in rir_channels
            ), "Invalid channel index in `rir_channels`."

        recording_rvb = self.recording.reverb_rir(
            rir_recording=rir_recording,
            normalize_output=normalize_output,
            early_only=early_only,
            affix_id=affix_id,
            rir_channels=rir_channels,
            room_rng_seed=room_rng_seed,
            source_rng_seed=source_rng_seed,
        )
        # Match the supervision's id (and it's underlying recording id).
        supervisions_rvb = [
            s.reverb_rir(
                affix_id=affix_id,
            )
            for s in self.supervisions
        ]

        return fastcopy(
            self,
            id=f"{self.id}_rvb" if affix_id else self.id,
            recording=recording_rvb,
            supervisions=supervisions_rvb,
        )

    def merge_supervisions(
        self,
        merge_policy: str = "delimiter",
        merge_channels: bool = True,
        custom_merge_fn: Optional[Callable[[str, Iterable[Any]], Any]] = None,
    ) -> "MultiCut":
        """
        Return a copy of the cut that has all of its supervisions merged into
        a single segment. The ``channel`` attribute of all the segments in this case
        will be set to the union of all channels. If ``merge_channels`` is set to ``False``,
        the supervisions will be merged into a single segment per channel group. The
        ``channel`` attribute will not change in this case.

        The new start is the start of the earliest superivion, and the new duration
        is a minimum spanning duration for all the supervisions. The text fields of
        all segments are concatenated with a whitespace.

        :param merge_policy: one of "keep_first" or "delimiter". If "keep_first", we
            keep only the first segment's field value, otherwise all string fields
            (including IDs) are prefixed with "cat#" and concatenated with a hash symbol "#".
            This is also applied to ``custom`` fields. Fields with a ``None`` value are omitted.
        :param merge_channels: If true, we will merge all supervisions into a single segment.
            If false, we will merge supervisions per channel group. Default: True.
        :param custom_merge_fn: a function that will be called to merge custom fields values.
            We expect ``custom_merge_fn`` to handle all possible custom keys.
            When not provided, we will treat all custom values as strings.
            It will be called roughly like:
            ``custom_merge_fn(custom_key, [s.custom[custom_key] for s in sups])``
        """
        merge_func_ = partial(
            merge_items_with_delimiter,
            delimiter="#",
            return_first=(merge_policy == "keep_first"),
        )

        # "m" stands for merged in variable names below

        if custom_merge_fn is not None:
            # Merge custom fields with the user-provided function.
            merge_custom = custom_merge_fn
        else:
            # Merge the string representations of custom fields.
            merge_custom = lambda k, vs: merge_func_(map(str, vs))

        sups = sorted(self.supervisions, key=lambda s: s.start)
        if len(sups) <= 1:
            return self

        if merge_channels:
            # Merge all supervisions into a single segment.
            all_channels = set()
            for s in sups:
                c = set(to_list(s.channel))
                all_channels.update(c)
            all_channels = sorted(all_channels)
            sups_by_channel = {tuple(all_channels): sups}  # `set` is not hashable
        else:
            # Merge supervisions per channel group.
            sups_by_channel = {
                tuple(c): list(csups)
                for c, csups in groupby(
                    sorted(sups, key=lambda s: s.channel), key=lambda s: s.channel
                )
            }

        msups = []
        text_overlap_warning = False
        for channel, csups in sups_by_channel.items():
            mstart = csups[0].start
            mend = csups[-1].end
            mduration = add_durations(mend, -mstart, sampling_rate=self.sampling_rate)

            custom_keys = set(
                k for s in csups if s.custom is not None for k in s.custom.keys()
            )
            alignment_keys = set(
                k for s in csups if s.alignment is not None for k in s.alignment.keys()
            )

            if (
                any(overlaps(s1, s2) for s1, s2 in zip(csups, csups[1:]))
                and any(s.text is not None for s in csups)
                and not text_overlap_warning
            ):
                warnings.warn(
                    "You are merging overlapping supervisions that have text transcripts. "
                    "The result is likely to be unusable if you are going to train speech "
                    f"recognition models (cut id: {self.id})."
                )
                text_overlap_warning = True

            msups.append(
                SupervisionSegment(
                    id=merge_func_(s.id for s in csups),
                    recording_id=csups[0].recording_id,
                    start=mstart,
                    duration=mduration,
                    channel=list(channel),
                    text=" ".join(s.text for s in csups if s.text),
                    speaker=merge_func_(s.speaker for s in csups if s.speaker),
                    language=merge_func_(s.language for s in csups if s.language),
                    gender=merge_func_(s.gender for s in csups if s.gender),
                    custom={
                        k: merge_custom(
                            k,
                            (
                                s.custom[k]
                                for s in csups
                                if s.custom is not None and k in s.custom
                            ),
                        )
                        for k in custom_keys
                    },
                    alignment={
                        # Concatenate the lists of alignment units.
                        k: reduce(
                            add,
                            (
                                s.alignment[k]
                                for s in csups
                                if s.alignment is not None and k in s.alignment
                            ),
                        )
                        for k in alignment_keys
                    },
                )
            )

        return fastcopy(self, supervisions=msups)

    def with_channels(self, channels: Union[List[int], int]) -> DataCut:
        """
        Select specified channels from this cut.
        Supports extending to other channels available in the underlying :class:`Recording`.
        If a single channel is provided, we'll return a :class:`~lhotse.cut.MonoCut`,
        otherwise we'll return a :class:`~lhotse.cut.MultiCut`.
        """
        channel_is_int = isinstance(channels, int)
        assert set([channels] if channel_is_int else channels).issubset(
            set(self.recording.channel_ids)
        ), f"Cannot select {channels=} because they are not a subset of {self.recording.channel_ids=}"

        mono = channel_is_int or len(channels) == 1
        if mono:
            from .mono import MonoCut

            if not channel_is_int:
                (channels,) = channels
            return MonoCut(
                id=f"{self.id}-{channels}",
                recording=self.recording,
                start=self.start,
                duration=self.duration,
                channel=channels,
                supervisions=[
                    fastcopy(s, channel=channels)
                    for s in self.supervisions
                    if is_equal_or_contains(s.channel, channels)
                ],
                custom=self.custom,
            )

        return fastcopy(self, channel=channels)

    @staticmethod
    def from_mono(*cuts: DataCut) -> "MultiCut":
        """
        Convert one or more MonoCut to a MultiCut. If multiple mono cuts are provided, they
        must match in all fields except the channel. Each cut must have a distinct channel.

        :param cuts: the input cut(s).
        :return: a MultiCut with a single track.
        """
        from .mono import MonoCut

        assert all(isinstance(c, MonoCut) for c in cuts), "All cuts must be MonoCuts"
        assert (
            sum(
                1 for _ in groupby(cuts, key=lambda c: (c.recording_id, c.start, c.end))
            )
            == 1
        ), "Cuts must match in all fields except channel"
        assert len(set(c.channel for c in cuts)) == len(
            cuts
        ), "All cuts must have a distinct channel"

        data = cuts[0].to_dict()
        data.pop("type")

        return MultiCut(
            **{
                **data,
                "channel": sorted([c.channel for c in cuts]),
                "supervisions": [s for c in cuts for s in c.supervisions],
            }
        )

    def to_mono(self, mono_downmix: bool = False) -> Union["DataCut", List["DataCut"]]:
        """
        Convert a MultiCut to either a list of MonoCuts (one per channel) or a single
        MonoCut obtained by downmixing all channels.

        :param mono_downmix: If true, we will downmix all channels into a single MonoCut.
            If false, we will return a list of MonoCuts, one per channel.
        :return: a list of MonoCuts or a single MonoCut.
        """
        from .mixed import MixedCut, MixTrack
        from .mono import MonoCut

        mono_cuts = [
            MonoCut(
                id=f"{self.id}-{channel}",
                recording=self.recording,
                start=self.start,
                duration=self.duration,
                channel=channel,
                supervisions=[
                    fastcopy(s, channel=channel)
                    for s in self.supervisions
                    if is_equal_or_contains(s.channel, channel)
                ],
                custom=self.custom,
            )
            for channel in to_list(self.channel)
        ]
        if not mono_downmix:
            return mono_cuts

        # Downmix the mono cuts into a single MixedCut.
        mixed_cut = MixedCut(
            id=self.id,
            tracks=[
                MixTrack(cut=mono_cut, offset=0.0, snr=None) for mono_cut in mono_cuts
            ],
        )
        return mixed_cut.to_mono()

    @staticmethod
    def from_dict(data: dict) -> "MultiCut":
        from lhotse.serialization import deserialize_custom_field

        # Remove "type" field if exists.
        data.pop("type", None)

        features = (
            Features.from_dict(data.pop("features")) if "features" in data else None
        )
        recording = (
            Recording.from_dict(data.pop("recording")) if "recording" in data else None
        )
        supervision_infos = data.pop("supervisions") if "supervisions" in data else []

        if "custom" in data:
            deserialize_custom_field(data["custom"])

        if "type" in data:
            data.pop("type")

        return MultiCut(
            **data,
            features=features,
            recording=recording,
            supervisions=[SupervisionSegment.from_dict(s) for s in supervision_infos],
        )



from typing import Dict, List, Optional, Tuple, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.speechcommands import (
    download_speechcommands,
    prepare_speechcommands,
)
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("speechcommands_version", type=str)
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def speechcommands(
    speechcommands_version: str,
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
):
    """Speech Commands v0.01 or v0.02 data preparation."""
    prepare_speechcommands(
        speechcommands_version=speechcommands_version,
        corpus_dir=corpus_dir,
        output_dir=output_dir,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("speechcommands_version", type=str)
@click.argument("target_dir", type=click.Path())
@click.option("--force-download", is_flag=True, default=False, help="Force download")
def speechcommands(
    speechcommands_version: str,
    target_dir: Pathlike,
    force_download: Optional[bool] = False,
):
    """Speech Commands v0.01 or v0.02 download."""
    download_speechcommands(
        speechcommands_version=speechcommands_version,
        target_dir=target_dir,
        force_download=force_download,
    )



import multiprocessing
import queue
import threading
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from functools import partial
from typing import Callable, Dict, Generator, Iterable, Optional, Type

from tqdm.auto import tqdm


def parallel_map(
    fn: Callable,
    *iterables: Iterable,
    num_jobs: int = 1,
    queue_size: int = 5000,
    threads: bool = False,
) -> Generator:
    """
    Works like Python's ``map``, but parallelizes the execution of ``fn`` over ``num_jobs``
    subprocesses or threads.

    Under the hood, it spawns ``num_jobs`` producer jobs that put their results on a queue.
    The current thread becomes a consumer thread and this generator yields items from the queue
    to the caller, as they become available.

    Example::

        >>> for root in parallel_map(math.sqrt, range(1000), num_jobs=4):
        ...     print(root)

    :param fn: function/callable to execute on each element.
    :param iterables: one of more iterables (one for each parameter of ``fn``).
    :param num_jobs: the number of parallel jobs.
    :param queue_size: max number of result items stored in memory.
        Decreasing this number might save more memory when the downstream processing is slower than
        the producer jobs.
    :param threads: whether to use threads instead of processes for producers (false by default).
    :return: a generator over results from ``fn`` applied to each item of ``iterables``.
    """
    thread = SubmitterThread(
        fn, *iterables, num_jobs=num_jobs, queue_size=queue_size, threads=threads
    )
    thread.start()
    q = thread.queue

    while thread.is_alive() or not q.empty():
        try:
            yield q.get(block=True, timeout=0.1).result()
        except queue.Empty:
            # Hit the timeout but thread is still running, try again.
            # This is needed to avoid hanging at the end when nothing else
            # shows up in the queue, but the thread didn't shutdown yet.
            continue

    thread.join()


class SubmitterThread(threading.Thread):
    def __init__(
        self,
        fn: Callable,
        *iterables,
        num_jobs: int = 1,
        queue_size: int = 10000,
        threads: bool = False,
    ) -> None:
        super().__init__()
        self.fn = fn
        self.num_jobs = num_jobs
        self.iterables = iterables
        self.queue = queue.Queue(maxsize=queue_size)
        self.use_threads = threads

    def run(self) -> None:
        executor = ThreadPoolExecutor if self.use_threads else ProcessPoolExecutor
        with executor(self.num_jobs) as ex:
            for args in zip(*self.iterables):
                future = ex.submit(self.fn, *args)
                self.queue.put(future, block=True)


class ParallelExecutor:
    """
    A class which uses ProcessPoolExecutor to parallelize the execution of a callable class.
    The instances of the runner class are instantiated separately in each worker process.

    Example::

        >>> class MyRunner:
        ...     def __init__(self):
        ...         self.name = name
        ...     def __call__(self, x):
        ...         return f'processed: {x}'
        ...
        >>> runner = ParallelExecutor(MyRunner, num_jobs=4)
        >>> for item in runner(range(10)):
        ...     print(item)

    If the __init__ method of the callable class accepts parameters except for `self`,
    use `functools.partial` or similar method to obtain a proper initialization function:

        >>> class MyRunner:
        ...     def __init__(self, name):
        ...         self.name = name
        ...     def __call__(self, x):
        ...         return f'{self.name}: {x}'
        ...
        >>> runner = ParallelExecutor(partial(MyRunner, name='my_name'), num_jobs=4)
        >>> for item in runner(range(10)):
        ...     print(item)


    The initialization function will be called separately for each worker process. Steps like loading a
    PyTorch model instance to the selected device should be done inside the initialization function.
    """

    _runners: Dict[Optional[int], Callable] = {}

    def __init__(
        self,
        init_fn: Callable[[], Callable],
        num_jobs: int = 1,
        verbose: bool = False,
        description: str = "Processing",
    ):
        """
        Instantiate a parallel executor.

        :param init_fn: A function which returns a runner object (e.g. a class) that will be instantiated
            in each worker process.
        :param num_jobs: The number of parallel jobs to run. Defaults to 1 (no parallelism).
        :param verbose: Whether to show a progress bar.
        :param description: The description to show in the progress bar.
        """
        self._make_runner = init_fn
        self.num_jobs = num_jobs
        self.verbose = verbose
        self.description = description

    def _init_runner(self):
        pid = multiprocessing.current_process().pid
        self._runners[pid] = self._make_runner()

    def _process(self, *args, **kwargs):
        pid = multiprocessing.current_process().pid
        runner = self._runners[pid]
        return runner(*args, **kwargs)

    def __call__(self, items: Iterable, **kwargs) -> Generator:
        if self.num_jobs == 1:
            runner = self._make_runner()
            for item in tqdm(items, desc=self.description, disable=not self.verbose):
                yield runner(item, **kwargs)

        else:
            pool = ProcessPoolExecutor(
                max_workers=self.num_jobs,
                initializer=self._init_runner,
                mp_context=multiprocessing.get_context("spawn"),
            )

            with pool as executor:
                try:
                    res = executor.map(partial(self._process, **kwargs), items)
                    for item in tqdm(
                        res,
                        desc=self.description,
                        total=len(items),
                        disable=not self.verbose,
                    ):
                        yield item
                except KeyboardInterrupt as exc:  # pragma: no cover
                    pool.shutdown(wait=False)
                    if self.verbose:
                        print("Interrupted by the user.")
                    raise exc
                except Exception as exc:  # pragma: no cover
                    pool.shutdown(wait=False)
                    raise RuntimeError(
                        "Parallel processing failed. Please report this issue."
                    ) from exc
                finally:
                    self._runners.clear()



"""
VoxConverse is an audio-visual diarisation dataset consisting of multispeaker clips of human speech, extracted from YouTube videos.
Updates and additional information about the dataset can be found at our website (https://www.robots.ox.ac.uk/~vgg/data/voxconverse/index.html).
"""

import json
import logging
import re
import shutil
import zipfile
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Dict, Iterable, Optional, Union
from urllib.error import HTTPError

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import AlignmentItem, SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download

DEV_AUDIO_ZIP = (
    "https://www.robots.ox.ac.uk/~vgg/data/voxconverse/data/voxconverse_dev_wav.zip"
)
TEST_AUDIO_ZIP = (
    "https://www.robots.ox.ac.uk/~vgg/data/voxconverse/data/voxconverse_test_wav.zip"
)
ANNOTATIONS_ZIP = "https://github.com/joonson/voxconverse/archive/master.zip"


def download_voxconverse(
    corpus_dir: Pathlike,
    force_download: bool = False,
):
    corpus_dir = Path(corpus_dir)
    corpus_dir.mkdir(parents=True, exist_ok=True)
    completed_detector = corpus_dir / ".completed"

    if not completed_detector.is_file() or force_download:
        print("Downloading VoxConverse dev set")
        resumable_download(DEV_AUDIO_ZIP, corpus_dir / "dev.zip")
        with zipfile.ZipFile(corpus_dir / "dev.zip") as zip_f:
            zip_f.extractall(corpus_dir / "dev")

        shutil.copytree(
            corpus_dir / "dev/audio", corpus_dir / "dev", dirs_exist_ok=True
        )
        shutil.rmtree(corpus_dir / "dev/audio")

        print("Downloading VoxConverse test set")
        resumable_download(TEST_AUDIO_ZIP, corpus_dir / "test.zip")
        with zipfile.ZipFile(corpus_dir / "test.zip") as zip_f:
            zip_f.extractall(corpus_dir / "test")

        shutil.copytree(
            corpus_dir / "test/voxconverse_test_wav",
            corpus_dir / "test",
            dirs_exist_ok=True,
        )
        shutil.rmtree(corpus_dir / "test/voxconverse_test_wav")

        print("Downloading VoxConverse annotations")
        resumable_download(ANNOTATIONS_ZIP, corpus_dir / "annotations.zip")
        with zipfile.ZipFile(corpus_dir / "annotations.zip") as zip_f:
            zip_f.extractall(corpus_dir)

        shutil.copytree(
            corpus_dir / "voxconverse-master", corpus_dir, dirs_exist_ok=True
        )
        shutil.rmtree(corpus_dir / "voxconverse-master")

        # cleanup
        (corpus_dir / "dev.zip").unlink()
        (corpus_dir / "test.zip").unlink()
        (corpus_dir / "annotations.zip").unlink()
        completed_detector.touch()

    print("Done")


def prepare_voxconverse(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    split_test: bool = False,  # test part is larger than dev part - split it into dev and test by default
):
    corpus_dir = Path(corpus_dir).absolute()

    splits = {}
    if split_test:
        splits["train"] = sorted((corpus_dir / "dev").glob("*.wav"))
        test_files = sorted((corpus_dir / "test").glob("*.wav"))
        splits["dev"] = test_files[: len(test_files) // 2]
        splits["test"] = test_files[len(test_files) // 2 :]
    else:
        splits["dev"] = sorted((corpus_dir / "dev").glob("*.wav"))
        splits["test"] = sorted((corpus_dir / "test").glob("*.wav"))

    manifests = {}
    for subset, wavs in splits.items():
        recordings = []
        supervisions = []
        for wav_file in wavs:
            recordings.append(Recording.from_file(wav_file))
            rttm_file = wav_file.with_suffix("").with_suffix(".rttm")
            for ix, (start, duration, speaker) in enumerate(_read_rttm(rttm_file)):
                supervisions.append(
                    SupervisionSegment(
                        id=f"{wav_file.stem}-{ix}",
                        recording_id=wav_file.stem,
                        start=start,
                        duration=duration,
                        channel=0,
                        language="en",
                        speaker=speaker,
                    )
                )

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            output_dir = Path(output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
            recording_set.to_file(
                output_dir / f"voxconverse_recordings_{subset}.jsonl.gz"
            )
            supervision_set.to_file(
                output_dir / f"voxconverse_supervisions_{subset}.jsonl.gz"
            )

        manifests[subset] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    return manifests


def _read_rttm(filename):
    with open(filename, "r") as f:
        for line in f.readlines():
            line = line.strip()
            if line.startswith("SPEAKER"):
                _, _, _, start, duration, _, _, speaker, _, _ = line.split()
                yield float(start), float(duration), speaker



import logging
import tarfile
from io import BytesIO
from typing import List, Optional

from lhotse.serialization import open_best


class TarWriter:
    """
    TarWriter is a convenience wrapper over :class:`tarfile.TarFile` that
    allows writing binary data into tar files that are automatically segmented.
    Each segment is a separate tar file called a "shard."

    Shards are useful in training of deep learning models that require a substantial
    amount of data. Each shard can be read sequentially, which allows faster reads
    from magnetic disks, NFS, or otherwise slow storage.

    Example::

        >>> with TarWriter("some_dir/data.%06d.tar", shard_size=100) as w:
        ...     w.write("blob1", binary_blob1)
        ...     w.write("blob2", binary_blob2)  # etc.

    It would create files such as ``some_dir/data.000000.tar``, ``some_dir/data.000001.tar``, etc.

    It's also possible to use ``TarWriter`` with automatic sharding disabled::

        >>> with TarWriter("some_dir/data.tar", shard_size=None) as w:
        ...     w.write("blob1", binary_blob1)
        ...     w.write("blob2", binary_blob2)  # etc.

    This class is heavily inspired by the WebDataset library:
    https://github.com/webdataset/webdataset
    """

    def __init__(self, pattern: str, shard_size: Optional[int] = 1000):
        self.pattern = str(pattern)
        if self.sharding_enabled and shard_size is None:
            raise RuntimeError(
                "shard_size must be specified when sharding is enabled via a formatting marker such as '%06d'"
            )
        if not self.sharding_enabled and shard_size is not None:
            logging.warning(
                "Sharding is disabled because `pattern` doesn't contain a formatting marker (e.g., '%06d'), "
                "but shard_size is not None - ignoring shard_size."
            )
        self.shard_size = shard_size
        self.gzip = self.pattern.endswith(".gz")
        self.reset()

    @property
    def sharding_enabled(self) -> bool:
        return "%" in self.pattern

    def reset(self):
        self.fname = None
        self.stream = None
        self.tarstream = None
        self.num_shards = 0
        self.num_items = 0
        self.num_items_total = 0

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def close(self):
        if self.tarstream is not None:
            self.tarstream.close()
        if self.stream is not None:
            self.stream.close()

    def _next_stream(self):
        self.close()

        if self.sharding_enabled:
            self.fname = self.pattern % self.num_shards
            self.num_shards += 1
        else:
            self.fname = self.pattern

        self.stream = open_best(self.fname, "wb")
        self.tarstream = tarfile.open(
            fileobj=self.stream, mode="w|gz" if self.gzip else "w|"
        )

        self.num_items = 0

    @property
    def output_paths(self) -> List[str]:
        if self.sharding_enabled:
            return [self.pattern % i for i in range(self.num_shards)]
        return [self.pattern]

    def write(self, key: str, data: BytesIO, count: bool = True):
        if count and (
            # the first item written
            self.num_items_total == 0
            or (
                # desired shard size achieved
                self.sharding_enabled
                and self.num_items > 0
                and self.num_items % self.shard_size == 0
            )
        ):
            self._next_stream()

        ti = tarfile.TarInfo(key)
        data.seek(0)
        ti.size = len(data.getvalue())
        self.tarstream.addfile(ti, data)
        if count:
            self.num_items += 1
            self.num_items_total += 1



from typing import List, Optional, Sequence, Tuple, Union

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.slu import prepare_slu
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path())
@click.argument("output_dir", type=click.Path())
def slu(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
):
    prepare_slu(corpus_dir=corpus_dir, output_dir=output_dir)



from typing import Optional

import click

from lhotse.bin.modes import prepare
from lhotse.recipes import prepare_eval2000
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus-dir", type=click.Path(exists=True, file_okay=False))
@click.argument("output-dir", type=click.Path())
@click.option(
    "--transcript-dir",
    type=click.Path(exists=True, file_okay=False),
    default=None,
    required=False,
)
@click.option(
    "--absolute-paths",
    default=False,
    help="Whether to return absolute or relative (to the corpus dir) paths for recordings.",
)
def eval2000(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    absolute_paths: bool,
    transcript_dir: Optional[Pathlike] = None,
):
    """
    The Eval2000 corpus preparation.

    \b
    This is conversational telephone speech collected as 2-channel, 8kHz-sampled data.
    The catalog number LDC2002S09 for audio corpora and LDC2002T43 for transcripts.

    This data is not available for free - your institution needs to have an LDC subscription.
    """

    prepare_eval2000(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
        absolute_paths=absolute_paths,
        transcript_path=transcript_dir,
    )



"""
This dataset consists of transcripts for 663 podcasts from the This American Life radio program from 1995 to 2020, covering 637 hours of audio (57.7 minutes per conversation) and an average of 18 unique speakers per conversation.

We hope that this dataset can serve as a new benchmark for the difficult tasks of speech transcription, speaker diarization, and dialog modeling on long, open-domain, multi-speaker conversations.

To learn more, please read our paper at: https://arxiv.org/pdf/2005.08072.pdf, and check the README.txt.
"""
import glob
import json
import logging
import re
import zipfile
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Dict, Iterable, Optional, Union
from urllib.error import HTTPError

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import AlignmentItem, SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download


def scrape_urls(website_url, output_path, year_range=(1995, 2021)):
    if not is_module_available("bs4"):
        raise ImportError("Please 'pip install beautifulsoup4' first.")

    import requests
    from bs4 import BeautifulSoup

    urls = {}
    for year in range(*year_range):
        print(f"Scraping {year}...")
        url = f"{website_url}/archive?year={year}"
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        page_urls = set()
        for a in soup.find_all("a", href=True, class_="goto-episode"):
            if a["href"].startswith("/"):
                page_urls.add(f"{website_url}{a['href']}")

        print(f"Found {len(page_urls)} episodes in {year}.")

        for episode_url in tqdm(page_urls):
            episode_id = int(episode_url.split("/")[-2])
            response = requests.get(episode_url)
            soup = BeautifulSoup(response.text, "html.parser")
            for a in soup.find_all("a", href=True, download=True):
                urls[f"ep-{episode_id}"] = a["href"]

    print(f"Saving results ({len(urls)} episodes)...")
    with open(output_path, "w") as f:
        json.dump(urls, f)


def included_episodes(target_dir: Pathlike) -> Iterable[str]:
    for subset in ["train", "valid", "test"]:
        with open(Path(target_dir) / f"{subset}-transcripts-aligned.json") as f:
            for episode_id in json.load(f).keys():
                yield episode_id


def download_this_american_life(
    target_dir: Pathlike = ".",
    force_download: bool = False,
    metadata_url="https://ipfs.io/ipfs/bafybeidyt3ch6t4dtu2ehdriod3jvuh34qu4pwjyoba2jrjpmqwckkr6q4/this_american_life.zip",
    website_url="https://thisamericanlife.org",
):
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    zip_path = target_dir / "metadata.zip"
    completed_detector = target_dir / "README.txt"

    if not completed_detector.is_file() or force_download:
        resumable_download(metadata_url, zip_path, force_download=force_download)

        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            print("Extracting...")
            zip_ref.extractall(target_dir)

        zip_path.unlink()

    # This American Life website was updated since the dataset annotations were published.
    # The links in the HTML page included are no longer valid and need to be re-scraped.
    urls_path = target_dir / "urls.json"
    if not urls_path.is_file():
        scrape_urls(website_url, urls_path)

    with open(urls_path) as f:
        urls = json.load(f)

    audio_dir = target_dir / "audio"
    audio_dir.mkdir(exist_ok=True)
    for ep_id in included_episodes(target_dir):
        print(f"Downloading episode {ep_id}... ({urls[ep_id]})")

        try:
            resumable_download(
                urls[ep_id], audio_dir / f"{ep_id}.mp3", force_download=force_download
            )
        except HTTPError as e:
            # Some episodes are no longer available on the website (like removed for anonymity reason, ep-374).
            print(f"Failed to download {ep_id}: {e}. Skipping...")
            continue

    print("Done!")


def prepare_this_american_life(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
):
    manifests = {}
    for subset in ["train", "dev", "test"]:
        manifests[subset] = prepare_this_american_life_subset(
            corpus_dir=corpus_dir,
            subset=subset,
            output_dir=output_dir,
        )

    return manifests


def prepare_this_american_life_subset(
    corpus_dir: Pathlike,
    subset: str,
    output_dir: Optional[Pathlike] = None,
):
    if not is_module_available("nltk"):
        raise ImportError("Please 'pip install nltk' first.")

    from nltk import word_tokenize

    corpus_dir = Path(corpus_dir).absolute()

    file_subset = "valid" if subset == "dev" else subset
    with open(Path(corpus_dir) / f"{file_subset}-transcripts-aligned.json") as f:
        transcripts = json.load(f)

    recordings = []
    supervisions = []
    pbar = tqdm(transcripts.items())
    for ep_id, transcript in pbar:
        pbar.set_description(desc=f"Processing {subset} subset ({ep_id})")
        audio_path = corpus_dir / "audio" / f"{ep_id}.mp3"
        if not audio_path.is_file():
            logging.warning(f"File {audio_path} not found - skipping.")
            continue

        recordings.append(Recording.from_file(audio_path))

        for utt_ix, utt in enumerate(transcript):
            text = utt["utterance"]
            words = word_tokenize(text)
            if len(words) != utt["n_words"]:
                logging.warning(
                    f"Transcript mismatch for {ep_id}-{utt_ix}: {utt['n_words']} words in the transcript, {len(words)} tokens in the text."
                )

            alignments = [
                AlignmentItem(words[int(ix)], start, end - start)
                for start, end, ix in utt["alignments"]
                if ix < len(words)
            ]
            segment = SupervisionSegment(
                id=f"{ep_id}-{utt_ix}",
                recording_id=ep_id,
                start=utt["utterance_start"],
                duration=utt["utterance_end"] - utt["utterance_start"],
                channel=0,
                text=text,
                language="en",
                speaker=utt["speaker"],
            )
            segment = segment.with_alignment("word", alignments)
            supervisions.append(segment)

    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)
    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        recording_set.to_file(
            output_dir / f"this-american-life_recordings_{subset}.jsonl.gz"
        )
        supervision_set.to_file(
            output_dir / f"this-american-life_supervisions_{subset}.jsonl.gz"
        )

    return {"recordings": recording_set, "supervisions": supervision_set}



from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

import numpy as np
import torch

from lhotse.features.base import FeatureExtractor, register_extractor
from lhotse.utils import (
    EPSILON,
    Seconds,
    asdict_nonull,
    compute_num_frames_from_samples,
    is_module_available,
)


def log_mel_spectrogram(
    audio: Union[np.ndarray, torch.Tensor],
    # large-v3 using 128 filters, others use 80
    n_mels: int = 80,
    n_fft: int = 400,
    hop_length: int = 160,
    sampling_rate: int = 16000,
    device: Optional[Union[str, torch.device]] = None,
):
    """
    From https://github.com/openai/whisper/blob/main/whisper/audio.py

    Compute the log-Mel spectrogram of

    Parameters
    ----------
    audio: Union[str, np.ndarray, torch.Tensor], shape = (*)
        The path to audio or either a NumPy array or Tensor containing the audio waveform in 16 kHz

    n_mels: int
        The number of Mel-frequency filters, only 80 is supported

    device: Optional[Union[str, torch.device]]
        If given, the audio tensor is moved to this device before STFT

    Returns
    -------
    torch.Tensor, shape = (n_frames, 80)
        A Tensor that contains the Mel spectrogram
    """
    if is_module_available("librosa"):
        import librosa
    else:
        raise ImportError(
            "Librosa is not installed. Please install librosa before using LibrosaFbank extractor."
        )
    if not torch.is_tensor(audio):
        audio = torch.from_numpy(audio)

    if device is not None:
        audio = audio.to(device)

    if len(audio.shape) == 2:
        if audio.shape[0] > 1:
            raise ValueError("Whisper Fbank works only with single-channel recordings.")
        else:
            audio = audio[0]
    assert (
        len(audio.shape) == 1
    ), f"Whisper Fbank works only with single-channel recordings (shape: {audio.shape})"

    window = torch.hann_window(n_fft).to(audio.device)
    stft = torch.stft(audio, n_fft, hop_length, window=window, return_complex=True)
    magnitudes = stft[..., :-1].abs() ** 2

    filters = librosa.filters.mel(sr=sampling_rate, n_fft=n_fft, n_mels=n_mels)
    filters = torch.from_numpy(filters).to(device)
    mel_spec = filters @ magnitudes

    log_spec = torch.clamp(mel_spec, min=1e-10).log10()
    log_spec = torch.maximum(log_spec, log_spec.max() - 8.0)
    log_spec = (log_spec + 4.0) / 4.0

    padding = compute_num_frames_from_samples(
        num_samples=len(audio),
        frame_shift=hop_length / sampling_rate,
        sampling_rate=sampling_rate,
    )
    if padding > log_spec.shape[1]:
        log_spec = torch.nn.functional.pad(
            log_spec, (0, padding - log_spec.shape[1]), mode="constant"
        )
    # change shape from 80, n_frames to n_frames,80
    log_spec = log_spec.transpose(0, 1)

    return log_spec


@dataclass
class WhisperFbankConfig:
    num_filters: int = 80
    device: str = "cpu"

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "WhisperFbankConfig":
        return WhisperFbankConfig(**data)


@register_extractor
class WhisperFbank(FeatureExtractor):
    name = "whisper-fbank"
    config_type = WhisperFbankConfig

    def __init__(self, config: Optional[WhisperFbankConfig] = None):
        super().__init__(config=config)
        self.sampling_rate = 16000
        self.hop_length = 160
        self.num_filters = self.config.num_filters

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    @property
    def frame_shift(self) -> Seconds:
        return self.hop_length / self.sampling_rate

    def to(self, device: str):
        self.config.device = device

    def feature_dim(self, sampling_rate: int) -> int:
        return self.num_filters

    def extract(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> Union[np.ndarray, torch.Tensor]:
        assert sampling_rate == self.sampling_rate, (
            f"Fbank was instantiated for sampling_rate "
            f"{self.sampling_rate}, but "
            f"sampling_rate={sampling_rate} was passed to extract(). "
            "Note you can use CutSet/RecordingSet.resample() to change the audio sampling rate."
        )

        is_numpy = False
        if not isinstance(samples, torch.Tensor):
            samples = torch.from_numpy(samples)
            is_numpy = True

        feats = log_mel_spectrogram(
            samples,
            n_mels=self.num_filters,
            device=self.device,
        )

        if is_numpy:
            return feats.cpu().numpy()
        else:
            return feats

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b),
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(np.exp(features)))



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes import download_voxpopuli, prepare_voxpopuli
from lhotse.recipes.voxpopuli import (
    LANGUAGES,
    LANGUAGES_V2,
    S2S_SRC_LANGUAGES,
    S2S_TGT_LANGUAGES,
)
from lhotse.utils import Pathlike

__all__ = ["voxpopuli"]


@prepare.command()
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--task",
    type=click.Choice(["asr", "s2s", "lm"]),
    default="asr",
    help="The task for which to prepare the VoxPopuli data.",
    show_default=True,
)
@click.option(
    "--lang",
    type=click.Choice(LANGUAGES + LANGUAGES_V2),
    default="en",
    help="The language to prepare (only used if task is asr or lm).",
    show_default=True,
)
@click.option(
    "--src-lang",
    type=click.Choice(S2S_SRC_LANGUAGES),
    default=None,
    help="The source language (only used if task is s2s).",
    show_default=True,
)
@click.option(
    "--tgt-lang",
    type=click.Choice(S2S_TGT_LANGUAGES),
    default=None,
    help="The target language (only used if task is s2s).",
    show_default=True,
)
@click.option(
    "--num-jobs",
    "-j",
    type=int,
    default=1,
    help="Number of parallel jobs (can provide small speed-ups).",
    show_default=True,
)
def voxpopuli(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    task: str,
    lang: str,
    src_lang: str,
    tgt_lang: str,
    num_jobs: int,
):
    """voxpopuli data preparation."""
    prepare_voxpopuli(
        corpus_dir,
        output_dir=output_dir,
        task=task,
        lang=lang,
        source_lang=src_lang,
        target_lang=tgt_lang,
    )


@download.command()
@click.argument("target_dir", type=click.Path())
@click.option(
    "--subset",
    type=click.Choice(["asr", "10k", "100k", "400k"] + LANGUAGES + LANGUAGES_V2),
    default="asr",
)
def voxpopuli(target_dir: Pathlike, subset: str):
    """voxpopuli download."""
    download_voxpopuli(target_dir, subset)



import logging
from typing import List, Optional, Union

from lhotse.cut import Cut
from lhotse.serialization import SequentialJsonlWriter


class JsonlShardWriter:
    """
    JsonlShardWriter writes Cuts or dicts into multiple JSONL file shards.
    The JSONL can be compressed with gzip if the file extension ends with ``.gz``.

    Example::

        >>> with JsonlShardWriter("some_dir/cuts.%06d.jsonl.gz", shard_size=100) as w:
        ...     for cut in ...:
        ...         w.write(cut)

    It would create files such as ``some_dir/cuts.000000.jsonl.gz``, ``some_dir/cuts.000001.jsonl.gz``, etc.

    See also: :class:`~lhotse.shar.writers.tar.TarWriter`
    """

    def __init__(self, pattern: str, shard_size: Optional[int] = 1000):
        self.pattern = pattern
        if not self.sharding_enabled and shard_size is not None:
            logging.warning(
                "Sharding is disabled because `pattern` doesn't contain a formatting marker (e.g., '%06d'), "
                "but shard_size is not None - ignoring shard_size."
            )
        self.shard_size = shard_size
        self.reset()

    @property
    def sharding_enabled(self) -> bool:
        return "%" in self.pattern

    def reset(self):
        self.fname = None
        self.stream = None
        self.num_shards = 0
        self.num_items = 0
        self.num_items_total = 0

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def close(self):
        if self.stream is not None:
            self.stream.close()

    def _next_stream(self):
        self.close()

        if self.sharding_enabled:
            self.fname = self.pattern % self.num_shards
            self.num_shards += 1
        else:
            self.fname = self.pattern

        self.stream = SequentialJsonlWriter(self.fname)

        self.num_items = 0

    @property
    def output_paths(self) -> List[str]:
        if self.sharding_enabled:
            return [self.pattern % i for i in range(self.num_shards)]
        return [self.pattern]

    def write(self, data: Union[Cut, dict], flush: bool = False) -> None:
        if (
            # the first item written
            self.num_items_total == 0
            or (
                # desired shard size achieved
                self.sharding_enabled
                and self.num_items > 0
                and self.num_items % self.shard_size == 0
            )
        ):
            self._next_stream()

        self.stream.write(data, flush=flush)
        self.num_items += 1
        self.num_items_total += 1

    def write_placeholder(self, cut_id: str, flush: bool = False) -> None:
        self.write({"cut_id": cut_id}, flush=flush)



from typing import Sequence, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes import download_but_reverb_db, prepare_but_reverb_db
from lhotse.utils import Pathlike

__all__ = ["but_reverb_db"]


@prepare.command()
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--parts",
    "-p",
    type=str,
    multiple=True,
    default=["silence", "rir"],
    show_default=True,
    help="Parts to prepare.",
)
def but_reverb_db(
    corpus_dir: Pathlike, output_dir: Pathlike, parts: Union[str, Sequence[str]]
):
    """BUT Reverb DB data preparation."""
    prepare_but_reverb_db(corpus_dir, output_dir=output_dir, parts=parts)


@download.command()
@click.argument("target_dir", type=click.Path())
@click.option(
    "--force-download",
    type=bool,
    default=False,
    help="If True, download even if file is present.",
)
def but_reverb_db(target_dir: Pathlike, force_download: bool):
    """BUT Reverb DB download."""
    download_but_reverb_db(target_dir, force_download=force_download)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.ali_meeting import download_ali_meeting, prepare_ali_meeting
from lhotse.utils import Pathlike

__all__ = ["ali_meeting"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--mic", type=click.Choice(["near", "far", "ihm", "sdm", "mdm"]), default="far"
)
@click.option(
    "--normalize-text",
    type=click.Choice(["none", "m2met"], case_sensitive=False),
    default="none",
    help="Type of text normalization to apply (M2MeT style is from the official challenge)",
)
@click.option(
    "--save-mono",
    is_flag=True,
    default=False,
    help="If True and `mic` is sdm, extract first channel and save as new recording.",
)
def ali_meeting(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    mic: str,
    normalize_text: str,
    save_mono: bool,
):
    """AliMeeting data preparation."""
    prepare_ali_meeting(
        corpus_dir,
        output_dir=output_dir,
        mic=mic,
        normalize_text=normalize_text,
        save_mono=save_mono,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option("--force-download", is_flag=True, default=False)
def ali_meeting(target_dir: Pathlike, force_download: bool):
    """AliMeeting download."""
    download_ali_meeting(target_dir, force_download=force_download)



import random
from copy import deepcopy
from functools import reduce
from operator import add
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np

from lhotse import CutSet
from lhotse.cut import Cut
from lhotse.dataset.sampling.base import CutSampler, SamplingDiagnostics
from lhotse.dataset.sampling.simple import SimpleCutSampler


class BucketingSampler(CutSampler):
    """
    Sorts the cuts in a :class:`CutSet` by their duration and puts them into similar duration buckets.
    For each bucket, it instantiates a simpler sampler instance, e.g. :class:`SimpleCutSampler`.

    It behaves like an iterable that yields lists of strings (cut IDs).
    During iteration, it randomly selects one of the buckets to yield the batch from,
    until all the underlying samplers are depleted (which means it's the end of an epoch).

    Examples:

    Bucketing sampler with 20 buckets, sampling single cuts::

        >>> sampler = BucketingSampler(
        ...    cuts,
        ...    # BucketingSampler specific args
        ...    sampler_type=SimpleCutSampler, num_buckets=20,
        ...    # Args passed into SimpleCutSampler
        ...    max_frames=20000
        ... )

    Bucketing sampler with 20 buckets, sampling pairs of source-target cuts::

        >>> sampler = BucketingSampler(
        ...    cuts, target_cuts,
        ...    # BucketingSampler specific args
        ...    sampler_type=CutPairsSampler, num_buckets=20,
        ...    # Args passed into CutPairsSampler
        ...    max_source_frames=20000, max_target_frames=15000
        ... )
    """

    def __init__(
        self,
        *cuts: CutSet,
        sampler_type: Type = SimpleCutSampler,
        num_buckets: int = 10,
        drop_last: bool = False,
        seed: int = 0,
        **kwargs: Any,
    ) -> None:
        """
        BucketingSampler's constructor.

        :param cuts: one or more ``CutSet`` objects.
            The first one will be used to determine the buckets for all of them.
            Then, all of them will be used to instantiate the per-bucket samplers.
        :param sampler_type: a sampler type that will be created for each underlying bucket.
        :param num_buckets: how many buckets to create.
        :param drop_last: When ``True``, we will drop all incomplete batches.
            A batch is considered incomplete if it depleted a bucket before
            hitting the constraint such as max_duration, max_cuts, etc.
        :param seed: random seed for bucket selection
        :param kwargs: Arguments used to create the underlying sampler for each bucket.
        """
        # Do not use the distributed capacities of the CutSampler in the top-level sampler.
        super().__init__(
            drop_last=drop_last,
            world_size=1,
            rank=0,
            seed=seed,
        )
        self.num_buckets = num_buckets
        self.sampler_type = sampler_type
        self.sampler_kwargs = kwargs
        self.cut_sets = cuts
        if any(cs.is_lazy for cs in self.cut_sets):
            raise ValueError(
                "BucketingSampler does not support working with lazy CutSet (e.g., "
                "those opened with 'load_manifest_lazy', 'CutSet.from_jsonl_lazy', or "
                "'CutSet.from_webdataset'). "
                "Please use lhotse.dataset.DynamicBucketingSampler instead."
            )

        # Split data into buckets.
        self.buckets = create_buckets_equal_duration(
            *self.cut_sets, num_buckets=num_buckets
        )

        # Create a separate sampler for each bucket.
        self.bucket_samplers = [
            self.sampler_type(
                *bucket_cut_sets,
                drop_last=drop_last,
                **self.sampler_kwargs,
            )
            for bucket_cut_sets in self.buckets
        ]

        # Initialize mutable state.
        self.bucket_rng = random.Random(self.seed + self.epoch)
        self.depleted = [False] * num_buckets

    @property
    def remaining_duration(self) -> Optional[float]:
        """
        Remaining duration of data left in the sampler (may be inexact due to float arithmetic).
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For BucketingSampler, it's the sum of remaining duration in all buckets.
        """
        try:
            return sum(
                s.remaining_duration for _, s in self._nondepleted_samplers_with_idxs
            )
        except TypeError:
            return None

    @property
    def remaining_cuts(self) -> Optional[int]:
        """
        Remaining number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For BucketingSampler, it's the sum of remaining cuts in all buckets.
        """
        try:
            return sum(
                s.remaining_cuts for _, s in self._nondepleted_samplers_with_idxs
            )
        except TypeError:
            return None

    @property
    def num_cuts(self) -> Optional[int]:
        """
        Total number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For BucketingSampler, it's the sum of num cuts in all buckets.
        """
        try:
            return sum(s.num_cuts for s in self.bucket_samplers)
        except TypeError:
            return None

    def set_epoch(self, epoch: int) -> None:
        """
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        :param epoch: Epoch number.
        """
        for s in self.bucket_samplers:
            s.set_epoch(epoch)
        super().set_epoch(epoch)

    def filter(self, predicate: Callable[[Cut], bool]) -> None:
        """
        Add a constraint on individual cuts that has to be satisfied to consider them.

        Can be useful when handling large, lazy manifests where it is not feasible to
        pre-filter them before instantiating the sampler.

        Example:
            >>> cuts = CutSet(...)
            ... sampler = SimpleCutSampler(cuts, max_duration=100.0)
            ... # Retain only the cuts that have at least 1s and at most 20s duration.
            ... sampler.filter(lambda cut: 1.0 <= cut.duration <= 20.0)
        """
        for sampler in self.bucket_samplers:
            sampler.filter(predicate)

    def allow_iter_to_reset_state(self):
        """
        Enables re-setting to the start of an epoch when iter() is called.
        This is only needed in one specific scenario: when we restored previous
        sampler state via ``sampler.load_state_dict()`` but want to discard
        the progress in the current epoch and start from the beginning.
        """
        super().allow_iter_to_reset_state()
        for s in self.bucket_samplers:
            s.allow_iter_to_reset_state()

    def state_dict(self) -> Dict[str, Any]:
        """
        Return the current state of the sampler in a state_dict.
        Together with ``load_state_dict()``, this can be used to restore the
        training loop's state to the one stored in the state_dict.
        """
        state_dict = super().state_dict()
        # We use deepcopies just in case somebody loads state dict during the same execution...
        state_dict.update(
            {
                "num_buckets": self.num_buckets,
                "depleted": deepcopy(self.depleted),
                "bucket_samplers": [s.state_dict() for s in self.bucket_samplers],
                "sampler_kwargs": deepcopy(self.sampler_kwargs),
                "bucket_rng_state": self.bucket_rng.getstate(),
            }
        )
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Restore the state of the sampler that is described in a state_dict.
        This will result in the sampler yielding batches from where the previous training left it off.

        .. caution::
            The samplers are expected to be initialized with the same CutSets,
            but this is not explicitly checked anywhere.

        .. caution::
            The input ``state_dict`` is being mutated: we remove each consumed key, and expect
            it to be empty at the end of loading. If you don't want this behavior, pass a copy
            inside of this function (e.g., using ``import deepcopy``).

        .. note::
            For implementers of sub-classes of CutSampler: the flag ``self._just_restored_state`` has to be
            handled in ``__iter__`` to make it avoid resetting the just-restored state (only once).
        """
        num_buckets = state_dict.pop("num_buckets")
        assert self.num_buckets == num_buckets, (
            "Error in BucketingSampler.load_state_dict(): Inconsistent number of buckets: "
            f"current sampler has {self.num_buckets}, the state_dict has {num_buckets}."
        )
        state_dict.pop("proportional_sampling", None)  # backward compatibility
        state_dict.pop("bucket_method", None)  # backward compatibility
        self.sampler_kwargs = state_dict.pop("sampler_kwargs")
        self.depleted = state_dict.pop("depleted")
        self.bucket_rng.setstate(state_dict.pop("bucket_rng_state"))

        assert len(self.bucket_samplers) == len(state_dict["bucket_samplers"]), (
            "Error in BucketingSampler.load_state_dict(): Inconsistent number of samplers: "
            f"current sampler has {len(self.bucket_samplers)}, "
            f"the state_dict has {len(state_dict['bucket_samplers'])}."
        )
        for sampler, sampler_sd in zip(
            self.bucket_samplers, state_dict.pop("bucket_samplers")
        ):
            sampler.load_state_dict(sampler_sd)

        super().load_state_dict(state_dict)

    def __iter__(self) -> "BucketingSampler":
        # Restored state with load_state_dict()? Skip resetting.
        if self._just_restored_state:
            return self
        # Why reset the current epoch?
        # Either we are iterating the epoch for the first time and it's a no-op,
        # or we are iterating the same epoch again, in which case setting more steps
        # than are actually available per epoch would have broken the checkpoint restoration.
        self.diagnostics.reset_current_epoch()
        # Reset the state to the beginning of the epoch.
        self.bucket_rng.seed(self.seed + self.epoch)
        for b in self.bucket_samplers:
            iter(b)
        self.depleted = [False] * self.num_buckets
        return self

    def _select_bucket_with_idx(self) -> Tuple[int, CutSampler]:
        if self.cut_sets[0].is_lazy:
            # With lazy CutSets, we simply choose a random bucket,
            # because we can't know how much data is left in the buckets.
            return self.bucket_rng.choice(self._nondepleted_samplers_with_idxs)
        idx_sampler_pairs = self._nondepleted_samplers_with_idxs
        if len(idx_sampler_pairs) == 1:
            # Only a single bucket left -- choose it.
            return idx_sampler_pairs[0]
        # If we got there, it means there are at least 2 buckets we can sample from.
        # We are going to use approximate proportional sampling:
        # for that, we randomly select two buckets, and then assign a higher probability
        # to the bucket that has more cumulative data duration left to sample.
        # This helps ensure that none of the buckets is depleted much earlier than
        # the others.
        idx1, sampler1 = self.bucket_rng.choice(idx_sampler_pairs)
        idx2, sampler2 = self.bucket_rng.choice(idx_sampler_pairs)
        # Note: prob1 is the probability of selecting sampler1
        try:
            prob1 = sampler1.remaining_duration / (
                sampler1.remaining_duration + sampler2.remaining_duration
            )
        except ZeroDivisionError:
            # This will happen when we have already depleted the samplers,
            # but the BucketingSampler doesn't know it yet. We only truly
            # know that a sampler is depleted when we try to get a batch
            # and it raises a StopIteration, which is done after this stage.
            # We can't depend on remaining_duration for lazy CutSets.
            # If both samplers are zero duration, just return the first one.
            return idx1, sampler1
        if self.bucket_rng.random() > prob1:
            return idx2, sampler2
        else:
            return idx1, sampler1

    def _next_batch(self):
        self.allow_iter_to_reset_state()
        while not self.is_depleted:
            idx, sampler = self._select_bucket_with_idx()
            try:
                return next(sampler)
            except StopIteration:
                self.depleted[idx] = True
        raise StopIteration()

    @property
    def is_depleted(self) -> bool:
        return all(self.depleted)

    @property
    def _nondepleted_samplers_with_idxs(self):
        return [
            (idx, bs)
            for idx, (bs, depleted) in enumerate(
                zip(self.bucket_samplers, self.depleted)
            )
            if not depleted
        ]

    def _log_diagnostics(self, batch: Union[CutSet, Tuple[CutSet, ...]]) -> None:
        return  # do nothing

    @property
    def diagnostics(self) -> SamplingDiagnostics:
        return reduce(add, (bucket.diagnostics for bucket in self.bucket_samplers))

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        return self.diagnostics.get_report()


def create_buckets_equal_duration(
    *cuts: CutSet, num_buckets: int
) -> List[Tuple[CutSet, ...]]:
    """
    Creates buckets of cuts with similar durations.
    Each bucket has the same cumulative duration, but a different number of cuts.

    :param cuts: One or more CutSets; the input CutSets are assumed to have the same cut IDs
        (i.e., the cuts correspond to each other and are meant to be sampled together as pairs,
        triples, etc.).
    :param num_buckets: The number of buckets.
    :return: A list of CutSet buckets (or tuples of CutSet buckets, depending on the input).
    """
    first_cut_set = cuts[0].sort_by_duration(ascending=True)
    buckets_per_cutset = [
        _create_buckets_equal_duration_single(first_cut_set, num_buckets=num_buckets)
    ]
    for cut_set in cuts[1:]:
        buckets_per_cutset.append(
            # .subset() will cause the output CutSet to have the same order of cuts as `bucket`
            cut_set.subset(cut_ids=bucket.ids)
            for bucket in buckets_per_cutset[0]
        )
    # zip(*buckets) does:
    # [(cs0_0, cs1_0, cs2_0), (cs0_1, cs1_1, cs2_1)] -> [(cs0_0, cs0_1), (cs1_0, cs1_1), (cs2_0, cs2_1)]
    return list(zip(*buckets_per_cutset))


def _create_buckets_equal_duration_single(
    cuts: CutSet, num_buckets: int
) -> List[CutSet]:
    """
    Helper method to partition a single CutSet into buckets that have the same
    cumulative duration.

    See also: :meth:`.create_buckets_from_duration_percentiles`.
    """
    total_duration = np.sum(c.duration for c in cuts)
    bucket_duration = total_duration / num_buckets
    # Define the order for adding cuts. We start at the beginning, then go to
    # the end, and work our way to the middle. Once in the middle we distribute
    # excess cuts among the two buckets close to the median duration. This
    # handles the problem of where to place cuts that caused previous buckets
    # to "over-flow" without sticking all of them in the last bucket, which
    # causes one large bucket at the end and also places many small duration
    # cuts with longer ones.
    order = list(range(0, len(cuts), 2)) + list(
        range(len(cuts) - (1 + len(cuts) % 2), 0, -2)
    )
    order2idx = {o_idx: i for i, o_idx in enumerate(order)}
    durations = [c.duration for c in cuts]

    # We need a list of the cut durations in the same order (0, N-1, 1, N-2, ...)
    ordered_cut_durations = sorted(zip(order, durations), key=lambda x: x[0])
    last_order, first_bucket = 0, 0
    last_bucket = num_buckets - 1
    buckets_dict = {i: 0 for i in range(num_buckets)}
    buckets_cut_dict = {i: [] for i in range(num_buckets)}
    middle_bucket = None
    idx_to_bucket_id = {}
    for i, (order_idx, duration) in enumerate(ordered_cut_durations, 1):
        # Check if we are at the middle bucket. first_bucket is the left bucket
        # we are processing. last_bucket is the right bucket. When they are the
        # same we are filling the bucket with cuts near the median duration.
        if middle_bucket is None and first_bucket == last_bucket:
            middle_bucket = first_bucket

        # i % 2 = 1 ==> process the left_bucket (first_bucket)
        if i % 2:
            if buckets_dict[first_bucket] + duration > bucket_duration:
                if middle_bucket is not None and first_bucket == middle_bucket:
                    first_bucket = max(0, min(middle_bucket - 1, num_buckets - 1))
                else:
                    first_bucket = min(first_bucket + 1, num_buckets - 1)
            buckets_dict[first_bucket] += duration
            idx_to_bucket_id[order2idx[order_idx]] = first_bucket
        # i % 2 = 0 ==> process the right bucket (last_bucket)
        else:
            if buckets_dict[last_bucket] + duration > bucket_duration:
                if middle_bucket is not None and last_bucket == middle_bucket:
                    last_bucket = max(middle_bucket + 1, 0)
                else:
                    last_bucket = max(last_bucket - 1, 0)
            buckets_dict[last_bucket] += duration
            idx_to_bucket_id[order2idx[order_idx]] = last_bucket

    # Now that buckets have been assigned, create the new cutset.
    for cut_idx, cut in enumerate(cuts):
        buckets_cut_dict[idx_to_bucket_id[cut_idx]].append(cut)
    buckets = [CutSet.from_cuts(buckets_cut_dict[i]) for i in range(num_buckets)]
    return buckets



from .adept import download_adept, prepare_adept
from .aishell import download_aishell, prepare_aishell
from .aishell3 import download_aishell3, prepare_aishell3
from .aishell4 import download_aishell4, prepare_aishell4
from .ali_meeting import download_ali_meeting, prepare_ali_meeting
from .ami import download_ami, prepare_ami
from .aspire import prepare_aspire
from .atcosim import download_atcosim, prepare_atcosim
from .babel import prepare_single_babel_language
from .baker_zh import download_baker_zh, prepare_baker_zh
from .bengaliai_speech import prepare_bengaliai_speech
from .broadcast_news import prepare_broadcast_news
from .but_reverb_db import download_but_reverb_db, prepare_but_reverb_db
from .bvcc import download_bvcc, prepare_bvcc
from .callhome_egyptian import prepare_callhome_egyptian
from .callhome_english import prepare_callhome_english
from .chime6 import download_chime6, prepare_chime6
from .cmu_arctic import download_cmu_arctic, prepare_cmu_arctic
from .cmu_indic import download_cmu_indic, prepare_cmu_indic
from .cmu_kids import prepare_cmu_kids
from .commonvoice import prepare_commonvoice
from .csj import prepare_csj
from .cslu_kids import prepare_cslu_kids
from .daily_talk import download_daily_talk, prepare_daily_talk
from .dihard3 import prepare_dihard3
from .dipco import download_dipco, prepare_dipco
from .earnings21 import download_earnings21, prepare_earnings21
from .earnings22 import download_earnings22, prepare_earnings22
from .edacc import download_edacc, prepare_edacc
from .eval2000 import prepare_eval2000
from .fisher_english import prepare_fisher_english
from .fisher_spanish import prepare_fisher_spanish
from .gale_arabic import prepare_gale_arabic
from .gale_mandarin import prepare_gale_mandarin
from .gigaspeech import prepare_gigaspeech
from .gigast import download_gigast, prepare_gigast
from .grid import download_grid, prepare_grid
from .heroico import download_heroico, prepare_heroico
from .hifitts import download_hifitts, prepare_hifitts
from .himia import download_himia, prepare_himia
from .icmcasr import prepare_icmcasr
from .icsi import download_icsi, prepare_icsi
from .iwslt22_ta import prepare_iwslt22_ta
from .kespeech import prepare_kespeech
from .l2_arctic import prepare_l2_arctic
from .libricss import download_libricss, prepare_libricss
from .librilight import prepare_librilight
from .librimix import download_librimix, prepare_librimix
from .librispeech import download_librispeech, prepare_librispeech
from .libritts import (
    download_libritts,
    download_librittsr,
    prepare_libritts,
    prepare_librittsr,
)
from .ljspeech import download_ljspeech, prepare_ljspeech
from .magicdata import download_magicdata, prepare_magicdata
from .mdcc import download_mdcc, prepare_mdcc
from .medical import download_medical, prepare_medical
from .mgb2 import prepare_mgb2
from .mls import prepare_mls
from .mobvoihotwords import download_mobvoihotwords, prepare_mobvoihotwords
from .mtedx import download_mtedx, prepare_mtedx
from .musan import download_musan, prepare_musan
from .nsc import prepare_nsc
from .peoples_speech import prepare_peoples_speech
from .reazonspeech import download_reazonspeech, prepare_reazonspeech
from .rir_noise import download_rir_noise, prepare_rir_noise
from .slu import prepare_slu
from .speechcommands import download_speechcommands, prepare_speechcommands
from .speechio import prepare_speechio
from .spgispeech import download_spgispeech, prepare_spgispeech
from .stcmds import download_stcmds, prepare_stcmds
from .switchboard import prepare_switchboard
from .tedlium import download_tedlium, prepare_tedlium
from .tedlium2 import download_tedlium2, prepare_tedlium2
from .thchs_30 import download_thchs_30, prepare_thchs_30
from .this_american_life import download_this_american_life, prepare_this_american_life
from .timit import download_timit, prepare_timit
from .uwb_atcc import download_uwb_atcc, prepare_uwb_atcc
from .vctk import download_vctk, prepare_vctk
from .voxceleb import download_voxceleb1, download_voxceleb2, prepare_voxceleb
from .voxconverse import download_voxconverse, prepare_voxconverse
from .voxpopuli import download_voxpopuli, prepare_voxpopuli
from .wenet_speech import prepare_wenet_speech
from .xbmu_amdo31 import download_xbmu_amdo31, prepare_xbmu_amdo31
from .yesno import download_yesno, prepare_yesno

__all__ = [
    "download_adept",
    "prepare_adept",
    "download_aishell",
    "prepare_aishell",
    "download_aishell3",
    "prepare_aishell3",
    "download_aishell4",
    "prepare_aishell4",
    "download_ali_meeting",
    "prepare_ali_meeting",
    "download_ami",
    "prepare_ami",
    "prepare_aspire",
    "download_atcosim",
    "prepare_atcosim",
    "prepare_single_babel_language",
    "prepare_bengaliai_speech",
    "prepare_broadcast_news",
    "download_but_reverb_db",
    "prepare_but_reverb_db",
    "download_bvcc",
    "prepare_bvcc",
    "prepare_callhome_egyptian",
    "prepare_callhome_english",
    "download_chime6",
    "prepare_chime6",
    "download_cmu_arctic",
    "prepare_cmu_arctic",
    "download_cmu_indic",
    "prepare_cmu_indic",
    "prepare_cmu_kids",
    "prepare_commonvoice",
    "prepare_csj",
    "prepare_cslu_kids",
    "download_daily_talk",
    "prepare_daily_talk",
    "prepare_dihard3",
    "download_dipco",
    "prepare_dipco",
    "download_earnings21",
    "prepare_earnings21",
    "download_earnings22",
    "prepare_earnings22",
    "download_edacc",
    "prepare_edacc",
    "prepare_eval2000",
    "prepare_fisher_english",
    "prepare_fisher_spanish",
    "prepare_gale_arabic",
    "prepare_gale_mandarin",
    "prepare_gigaspeech",
    "download_gigast",
    "prepare_gigast",
    "download_grid",
    "prepare_grid",
    "download_heroico",
    "prepare_heroico",
    "download_hifitts",
    "prepare_hifitts",
    "download_himia",
    "prepare_himia",
    "prepare_icmcasr",
    "download_icsi",
    "prepare_icsi",
    "prepare_iwslt22_ta",
    "prepare_kespeech",
    "prepare_l2_arctic",
    "download_libricss",
    "prepare_libricss",
    "prepare_librilight",
    "download_librimix",
    "prepare_librimix",
    "download_librispeech",
    "prepare_librispeech",
    "download_libritts",
    "download_librittsr",
    "prepare_libritts",
    "prepare_librittsr",
    "download_ljspeech",
    "prepare_ljspeech",
    "download_magicdata",
    "prepare_magicdata",
    "download_medical",
    "prepare_medical",
    "prepare_mgb2",
    "prepare_mls",
    "download_mobvoihotwords",
    "prepare_mobvoihotwords",
    "download_mtedx",
    "prepare_mtedx",
    "download_musan",
    "prepare_musan",
    "prepare_nsc",
    "prepare_peoples_speech",
    "download_reazonspeech",
    "prepare_reazonspeech",
    "download_rir_noise",
    "prepare_rir_noise",
    "prepare_slu",
    "download_speechcommands",
    "prepare_speechcommands",
    "download_spgispeech",
    "prepare_spgispeech",
    "download_stcmds",
    "prepare_stcmds",
    "prepare_switchboard",
    "download_tedlium",
    "prepare_tedlium",
    "download_thchs_30",
    "prepare_thchs_30",
    "download_this_american_life",
    "prepare_this_american_life",
    "download_timit",
    "prepare_timit",
    "download_uwb_atcc",
    "prepare_uwb_atcc",
    "download_vctk",
    "prepare_vctk",
    "download_voxceleb1",
    "download_voxceleb2",
    "prepare_voxceleb",
    "download_voxconverse",
    "prepare_voxconverse",
    "download_voxpopuli",
    "prepare_voxpopuli",
    "prepare_wenet_speech",
    "download_xbmu_amdo31",
    "prepare_xbmu_amdo31",
    "download_yesno",
    "prepare_yesno",
]



import logging
import os
import sys
from pathlib import Path


def default_tools_cachedir(force_mkdir: bool = False) -> Path:
    d = Path.home() / ".lhotse/tools"
    try:
        d.mkdir(exist_ok=True, parents=True)
    except OSError:
        if force_mkdir:
            raise
        else:
            logging.warning(
                f"We couldn't create lhotse utilities directory: {d} (not enough space/no permissions?)"
            )
    return d


def add_tools_to_path():
    sph2pipe_path = str(default_tools_cachedir() / "sph2pipe-2.5")
    sys.path.append(sph2pipe_path)
    os.environ["PATH"] += os.pathsep + sph2pipe_path  # platform-agnostic



import warnings
from dataclasses import asdict, dataclass
from typing import Optional, Tuple, Union

import numpy as np
import torch

from lhotse.augmentation.transform import AudioTransform
from lhotse.utils import EPSILON, Seconds, is_module_available


@dataclass
class LoudnessNormalization(AudioTransform):
    """
    Loudness normalization based on pyloudnorm: https://github.com/csteinmetz1/pyloudnorm.
    """

    target: float

    def __call__(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> np.ndarray:
        if torch.is_tensor(samples):
            samples = samples.cpu().numpy()
        augmented = normalize_loudness(
            samples, target=self.target, sampling_rate=sampling_rate
        )
        return augmented

    def reverse_timestamps(
        self, offset: Seconds, duration: Optional[Seconds], sampling_rate: int
    ) -> Tuple[Seconds, Optional[Seconds]]:
        return offset, duration


def normalize_loudness(
    audio: np.ndarray,
    target: float,
    sampling_rate: int = 16000,
) -> np.ndarray:
    """
    Applies pyloudnorm based loudness normalization to the input audio. The input audio
    can have up to 5 channels, with the following order: [left, right, center, left_surround, right_surround]

    :param audio: the input audio, expected to be 2D with shape (channels, samples).
    :param target: the target loudness in LUFS.
    :param sampling_rate: the sampling rate of the audio.
    :return: the loudness normalized audio.
    """
    if not is_module_available("pyloudnorm"):
        raise ImportError(
            "Please install pyloudnorm first using 'pip install pyloudnorm'"
        )

    import pyloudnorm as pyln

    assert audio.ndim == 2, f"Expected 2D audio shape, got: {audio.shape}"

    duration = audio.shape[1] / sampling_rate

    # measure the loudness first
    meter = pyln.Meter(
        sampling_rate, block_size=min(0.4, duration - EPSILON)
    )  # create BS.1770 meter
    loudness = meter.integrated_loudness(audio.T)

    # loudness normalize audio to target LUFS. We will ignore the warnings related to
    # clipping the audio.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        loudness_normalized_audio = pyln.normalize.loudness(audio.T, loudness, target)

    return loudness_normalized_audio.T



from collections import defaultdict
from typing import Callable, Dict, List, Union

import numpy as np
import torch
import torch.nn.functional as F

from lhotse import validate
from lhotse.cut import CutSet
from lhotse.dataset.input_strategies import BatchIO, PrecomputedFeatures
from lhotse.utils import LOG_EPSILON, compute_num_frames, ifnone
from lhotse.workarounds import Hdf5MemoryIssueFix


class K2SurtDataset(torch.utils.data.Dataset):
    """
    The PyTorch Dataset for the multi-talker ASR task using k2 library.
    We support the modeling framework known as Streaming Unmixing and Recognition
    Transducer (SURT), as described in [1] and [2], but this dataset can also be
    used for other multi-talker ASR approaches such as MT-RNNT [3] and SOT [4].
    See icefall recipe for usage: https://github.com/k2-fsa/icefall/pull/1126.

    We take a cut containing possibly overlapping speech and split the supervision
    segments into one of N channels based on their start times (N is provided), known
    as ``heuristic error assignment training`` (HEAT) [1]. The supervision segments
    in each channel are then concatenated and used as the supervision for that channel.
    If we have features for the source cuts, we can also return them for use in masking
    losses, for instance.

    [1] Lu, L., Kanda, N., Li, J., & Gong, Y. (2021). Streaming end-to-end multi-talker
    speech recognition. IEEE Signal Processing Letters, 28, 803-807.

    [2] Raj, D., Lu, L., Chen, Z., Gaur, Y., & Li, J. (2022, May). Continuous streaming
    multi-talker asr with dual-path transducers. In ICASSP 2022-2022 IEEE International
    Conference on Acoustics, Speech and Signal Processing (ICASSP) (pp. 7317-7321). IEEE.

    [3] Sklyar, I., Piunova, A., Zheng, X., & Liu, Y. (2022, May). Multi-turn RNN-T for
    streaming recognition of multi-party speech. In ICASSP 2022-2022 IEEE International
    Conference on Acoustics, Speech and Signal Processing (ICASSP) (pp. 8402-8406). IEEE.

    [4] Kanda, N., Gaur, Y., Wang, X., Meng, Z., & Yoshioka, T. (2020). Serialized output
    training for end-to-end overlapped speech recognition. arXiv preprint arXiv:2003.12687.

    .. hint:: Training mixtures can be simulated from single-speaker utterances using
        :class:`~lhotse.workflows.MeetingSimulation` workflow.

    This dataset expects to be queried with lists of cut IDs,
    for which it loads features and automatically collates/batches them.

    To use it with a PyTorch DataLoader, set ``batch_size=None``
    and provide a :class:`SimpleCutSampler` sampler.

    Each item in this dataset is a dict of:

    .. code-block::

        {
            'inputs': float tensor with shape determined by :attr:`input_strategy`:
                      - single-channel:
                        - features: (B, T, F)
                        - audio: (B, T)
                      - multi-channel: currently not supported
            'input_lens': int tensor of shape (B,)
            'supervisions': list of lists of supervision segments, where the outer list is
                        batch, and the inner list is indexed by channel. So ``len(supervisions) == B``,
                        and ``len(supervisions[i]) == num_channels``. Note that some channels may
                        have no supervision segments.
            'text': list of lists of strings, where the outer list is batch, and the inner list
                    is indexed by channel. So ``len(text) == B``, and ``len(text[i]) == num_channels``.
                    Each element contains the text of the supervision segments in that channel,
                    joined by the :attr:`text_delimiter`. Note that some channels may have no
                    supervision segments, so the corresponding text will be an empty string.
        }

    If ``return_cuts`` is ``True``, each item will also contain a ``cuts`` field with the
    :class:`~lhotse.cut.Cut` objects used to create the batch.

    Additionally, if ``return_sources`` is ``True``, each item will contain:

    .. code-block::

        {
            'source_feats': list of list of float tensors. The outer list is batch, and the inner
                            list is number of segments in the mixture. Each element denotes
                            the features of the source cut in the mixture.
            'source_boundaries': list of list of tuples, where the outer list is batch, and the inner list
                                is number of segments in the mixture. Each element denotes
                                the start and end frame of the source cut in the mixture.
        }

    In order to return the source features and boundaries, we expect the cuts to contain
    some additional fields:

    - ``source_feats``: a float tensor representing the features of all the source segments,
        concatenated together along T dimension, in order of their start times.
    - ``source_feat_offsets``: a list of ints, where each element denotes the offset of the
        source segments in the ``source_feats`` tensor.

    See https://github.com/lhotse-speech/lhotse/discussions/1008#discussioncomment-5511746
    for example code to create these fields.


    Dimension symbols legend:
    * ``B`` - batch size (number of Cuts)
    * ``S`` - number of supervision segments (greater or equal to B, as each Cut may have multiple supervisions)
    * ``T`` - number of frames of the longest Cut
    * ``F`` - number of features

    The 'sequence_idx' field is the index of the Cut used to create the example in the Dataset.
    """

    def __init__(
        self,
        return_cuts: bool = False,
        return_sources: bool = False,
        return_alignments: bool = False,
        num_channels: int = 2,
        text_delimiter: str = " ",
        cut_transforms: List[Callable[[CutSet], CutSet]] = None,
        input_transforms: List[Callable[[torch.Tensor], torch.Tensor]] = None,
        input_strategy: BatchIO = PrecomputedFeatures(),
        pad_value: float = LOG_EPSILON,
        strict: bool = False,
    ):
        """
        k2 ASR IterableDataset constructor.

        :param return_cuts: When ``True``, will additionally return a "cut" field in each batch with the Cut
            objects used to create that batch.
        :param return_sources: When ``True``, will additionally return a "source_feats" field and a "source_boundaries"
            field in each batch. The "source_feats" field contains the features of the source cuts from
            which the mixture was created, and "source_boundaries" contains the boundaries of the source cuts
            in the mixture (in number of frames). This requires that the cuts contain additional fields
            ``source_feats`` (which is a TemporalArray) and ``source_feat_offsets`` (which is a list of ints).
        :param return_alignments: When ``True``, will keep the supervision alignments if they
            are present in the cuts.
        :param num_channels: Number of output branches. The supervision utterances will be
            split into the channels based on their start times.
        :param text_delimiter: The delimiter used to join the text of the supervision segments in
            each channel.
        :param cut_transforms: A list of transforms to be applied on each sampled batch,
            before converting cuts to an input representation (audio/features).
            Examples: cut concatenation, noise cuts mixing, etc.
        :param input_transforms: A list of transforms to be applied on each sampled batch,
            after the cuts are converted to audio/features.
            Examples: normalization, SpecAugment, etc.
        :param input_strategy: The strategy used to convert the cuts to audio/features.
        :param pad_value: The value used to pad the source features to resolve one-off errors.
        :param strict: If ``True``, we will remove cuts that have more simultaneous supervisions
            than the number of channels. If ``False``, we will keep them.
        """
        super().__init__()
        # Initialize the fields
        self.return_cuts = return_cuts
        self.return_sources = return_sources
        self.return_alignments = return_alignments
        self.num_channels = num_channels
        self.text_delimiter = text_delimiter
        self.cut_transforms = ifnone(cut_transforms, [])
        self.input_transforms = ifnone(input_transforms, [])
        self.input_strategy = input_strategy
        self.pad_value = pad_value
        self.strict = strict

        # This attribute is a workaround to constantly growing HDF5 memory
        # throughout the epoch. It regularly closes open file handles to
        # reset the internal HDF5 caches.
        self.hdf5_fix = Hdf5MemoryIssueFix(reset_interval=100)

    def __getitem__(self, cuts: CutSet) -> Dict[str, Union[torch.Tensor, List[str]]]:
        """
        Return a new batch, with the batch size automatically determined using the constraints
        of max_frames and max_cuts.
        """
        validate_for_asr(cuts)

        self.hdf5_fix.update()

        if not self.return_alignments:
            cuts = cuts.drop_alignments()

        # Sort the cuts by duration so that the first one determines the batch time dimensions.
        cuts = cuts.sort_by_duration(ascending=False)

        # Optional CutSet transforms - e.g. padding, or speed perturbation that adjusts
        # the supervision boundaries.
        for tnfm in self.cut_transforms:
            cuts = tnfm(cuts)

        # Assign supervisions to channels based on their start times.
        # ``supervisions`` is a dict indexed by cut id, and each value is a list of
        # lists of supervisions. The outer list is indexed by channel, and the inner
        # list contains the supervisions for that channel.
        supervisions = defaultdict(list)
        invalid_cuts = []
        source_feats = []
        source_boundaries = []

        for cut in cuts:
            cut_sups = [[] for _ in range(self.num_channels)]
            last_sup_end = [0.0 for _ in range(self.num_channels)]

            cut_sources = []
            cut_source_boundaries = []
            invalid_cut = False

            for sup in sorted(cut.supervisions, key=lambda s: s.start):
                # Assign the supervision to the first channel that is either empty or
                # has a supervision that ends before the current supervision starts.
                assigned = False
                for i in range(self.num_channels):
                    if len(cut_sups[i]) == 0 or last_sup_end[i] <= sup.start:
                        cut_sups[i].append(sup)
                        last_sup_end[i] = max(last_sup_end[i], sup.end)
                        assigned = True
                        break

                if not assigned:
                    # If we reach here, it means that there is no channel that is empty
                    # or has a supervision that ends before the current supervision starts.
                    # This is possible if number of overlapping speakers is more than the
                    # number of available channels. In this case, we assign the supervision
                    # so as to minimize the overlapping part. For this, we select the
                    # channel which ends the earliest.
                    invalid_cut = True
                    min_end_channel = last_sup_end.index(min(last_sup_end))
                    cut_sups[min_end_channel].append(sup)
                    last_sup_end[min_end_channel] = max(
                        last_sup_end[min_end_channel], sup.end
                    )

            if self.return_sources:
                source_feat_offsets = cut.source_feat_offsets
                assert len(source_feat_offsets) == len(cut.supervisions), (
                    "The number of source feature offsets should be equal to the number of supervisions."
                    f"Got {len(source_feat_offsets)} offsets for {len(cut.supervisions)} supervisions."
                )
                # To get cut_sources, we split the source_feats into a list of tensors
                # based on the source_feat_offsets.
                cut_sources = [
                    torch.from_numpy(x)
                    for x in np.split(cut.load_source_feats(), source_feat_offsets[1:])
                ]
                # To get cut_source_boundaries, we create (start, end) tuples based on
                # the supervision start and end times.
                cut_source_boundaries = [
                    (
                        compute_num_frames(
                            sup.start, cut.frame_shift, cut.sampling_rate
                        ),
                        compute_num_frames(sup.end, cut.frame_shift, cut.sampling_rate),
                    )
                    for sup in sorted(
                        cut.supervisions, key=lambda s: (s.start, s.speaker)
                    )
                ]
                # Adjust the source feats to fix one-off errors in the source_feat_offsets.
                cut_sources = [
                    adjust_source_feats(x, end - start, padding_value=self.pad_value)
                    for x, (start, end) in zip(cut_sources, cut_source_boundaries)
                ]

            if invalid_cut and self.strict:
                invalid_cuts.append(cut.id)
                continue
            supervisions[cut.id] = cut_sups
            if self.return_sources:
                source_feats.append(cut_sources)
                source_boundaries.append(cut_source_boundaries)

        # Remove invalid cuts.
        if len(invalid_cuts) > 0:
            print(
                f"WARNING: {len(invalid_cuts)} cuts were removed out of {len(cuts)} due to more overlapping speakers than channels."
            )
            cuts = cuts.filter(lambda cut: cut.id not in invalid_cuts)

        # Get a tensor with batched feature matrices, shape (B, T, F)
        # Collation performs auto-padding, if necessary.
        input_tpl = self.input_strategy(cuts)
        if len(input_tpl) == 3:
            # An input strategy with fault tolerant audio reading mode.
            # "cuts" may be a subset of the original "cuts" variable,
            # that only has cuts for which we succesfully read the audio.
            inputs, input_lens, cuts = input_tpl
        else:
            inputs, input_lens = input_tpl

        batch = {
            "inputs": inputs,
            "input_lens": input_lens,
            "supervisions": list(supervisions.values()),
            "text": [
                [
                    self.text_delimiter.join([sup.text.strip() for sup in sups_ch])
                    for sups_ch in cut_sups
                ]
                for cut_sups in supervisions.values()
            ],
        }
        if self.return_cuts:
            batch["cuts"] = cuts
        if self.return_sources:
            batch["source_feats"] = source_feats
            batch["source_boundaries"] = source_boundaries
        return batch


def adjust_source_feats(feats, num_frames, padding_value=0.0, tol=2):
    """
    Adjust the number of frames in the source features to match the supervision.
    If the source features have fewer frames than the supervision, we pad them
    to match the supervision. If the source features have more frames than the
    supervision, we trim them to match the supervision.

    Args:
        feats: Source features.
        num_frames: Number of frames in the supervision.
        padding_value: Value to use for padding.
        tol: Tolerance for checking if the number of frames in the source features
            is close to the number of frames in the supervision.
    """
    if feats.shape[0] == num_frames:
        return feats
    elif abs(feats.shape[0] - num_frames) > tol:
        raise ValueError(
            f"Number of frames in the source features ({feats.shape[0]}) is not close to "
            f"the number of frames in the supervision ({num_frames})."
        )
    elif feats.shape[0] < num_frames:
        # If the source features have fewer frames than the supervision,
        # we pad them to match the supervision.
        return F.pad(feats, (0, 0, 0, num_frames - feats.shape[0]), value=padding_value)
    else:
        # If the source features have more frames than the supervision,
        # we trim them to match the supervision.
        return feats[:num_frames]


def validate_for_asr(cuts: CutSet) -> None:
    validate(cuts)
    tol = 2e-3  # 1ms
    for cut in cuts:
        for supervision in cut.supervisions:
            assert supervision.start >= -tol, (
                f"Supervisions starting before the cut are not supported for ASR"
                f" (sup id: {supervision.id}, cut id: {cut.id})"
            )
            assert supervision.duration <= cut.duration + tol, (
                f"Supervisions ending after the cut "
                f"are not supported for ASR"
                f" (sup id: {supervision.id}, cut id: {cut.id})"
            )



import click

from .cli_base import cli


@cli.command()
def list_audio_backends():
    """
    List the names of all available audio backends.
    """
    from lhotse import available_audio_backends

    click.echo(available_audio_backends())



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.stcmds import download_stcmds, prepare_stcmds
from lhotse.utils import Pathlike

__all__ = ["stcmds"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def stcmds(corpus_dir: Pathlike, output_dir: Pathlike):
    """Stcmds ASR data preparation."""
    prepare_stcmds(corpus_dir, output_dir=output_dir)


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def stcmds(target_dir: Pathlike):
    """Stcmds download."""
    download_stcmds(target_dir)



import math
from typing import Any, Dict, List, Optional

import numpy as np
import torch
from torch.utils.data import IterableDataset

from lhotse import RecordingSet, Seconds, compute_num_samples, validate
from lhotse.audio.backend import torchaudio_supports_ffmpeg
from lhotse.audio.utils import suppress_audio_loading_errors
from lhotse.augmentation import AugmentFn
from lhotse.cut import CutSet
from lhotse.dataset.collation import collate_audio, collate_features, collate_matrices
from lhotse.features import FeatureExtractor
from lhotse.features.kaldi.layers import _get_strided_batch_streaming


class UnsupervisedDataset(torch.utils.data.Dataset):
    """
    Dataset that contains no supervision - it only provides the features extracted from recordings.

    .. code-block::

        {
            'features': (B x T x F) tensor
            'features_lens': (B, ) tensor
        }
    """

    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, cuts: CutSet) -> Dict[str, Any]:
        self._validate(cuts)
        features, features_lens = collate_features(cuts)
        return {
            "cuts": cuts,
            "features": features,
            "features_lens": features_lens,
        }

    def _validate(self, cuts: CutSet) -> None:
        validate(cuts)
        assert all(cut.has_features for cut in cuts)


class UnsupervisedWaveformDataset(UnsupervisedDataset):
    """
    A variant of UnsupervisedDataset that provides waveform samples instead of features.
    The output is a tensor of shape (C, T), with C being the number of channels and T the number of audio samples.
    In this implementation, there will always be a single channel.

    Returns:

    .. code-block::

        {
            'audio': (B x NumSamples) float tensor
            'audio_lens': (B, ) int tensor
        }
    """

    def __init__(self, collate: bool = True) -> None:
        super().__init__()
        self.collate = collate

    def __getitem__(self, cuts: CutSet) -> Dict[str, Any]:
        if self.collate:
            audio, audio_lens = collate_audio(cuts)
            return {
                "cuts": cuts,
                "audio": audio,
                "audio_lens": audio_lens,
            }
        else:
            remain_cuts = []
            remain_audios = []
            for c in cuts:
                with suppress_audio_loading_errors():
                    remain_audios.append(c.load_audio())
                    remain_cuts.append(c)
            return {"cuts": CutSet.from_cuts(remain_cuts), "audio": remain_audios}

    def _validate(self, cuts: CutSet) -> None:
        validate(cuts)
        assert all(cut.has_recording for cut in cuts)


class DynamicUnsupervisedDataset(UnsupervisedDataset):
    """
    An example dataset that shows how to use on-the-fly feature extraction in Lhotse.
    It accepts two additional inputs - a FeatureExtractor and an optional WavAugmenter for time-domain data augmentation..
    The output is approximately the same as that of the ``UnsupervisedDataset`` -
    there might be slight differences for ``MixedCut``s, because this dataset mixes them in the time domain,
    and ``UnsupervisedDataset`` does that in the feature domain.
    Cuts that are not mixed will yield identical results in both dataset classes.
    """

    def __init__(
        self,
        feature_extractor: FeatureExtractor,
        augment_fn: Optional[AugmentFn] = None,
    ):
        super().__init__()
        self.feature_extractor = feature_extractor
        self.augment_fn = augment_fn

    def __getitem__(self, cuts: CutSet) -> torch.Tensor:
        self._validate(cuts)

        def generate_cut(cuts: CutSet):
            for cut in cuts:
                with suppress_audio_loading_errors():
                    yield cut.compute_features(
                        extractor=self.feature_extractor,
                        augment_fn=self.augment_fn,
                    )

        features = collate_matrices(generate_cut(cuts))
        return features

    def _validate(self, cuts: CutSet) -> None:
        validate(cuts)
        assert all(cut.has_recording for cut in cuts)


class RecordingChunkIterableDataset(IterableDataset):
    """
    This dataset iterates over chunks of a recording, for each recording provided.
    It supports setting a chunk_shift < chunk_size to run model predictions on
    overlapping audio chunks.

    The format of yielded items is the following::

        {
            "recording_id": str
            "begin_time": tensor with dtype=float32 shape=(1,)
            "end_time": tensor with dtype=float32 shape=(1,)
            "audio": tensor with dtype=float32 shape=(chunk_size_in_samples,)
        }

    Unlike most other datasets in Lhotse, this dataset does not yield batched items,
    and should be used like the following::

        >>> recordings = RecordingSet.from_file("my-recordings.jsonl.gz")
        ... dataset = RecordingChunkIterableDataset(recordings, chunk_size=30.0, chunk_shift=25.0)
        ... dloader = torch.utils.data.DataLoader(
        ...     dataset,
        ...     batch_size=32,
        ...     collate_fn=audio_chunk_collate,
        ...     worker_init_fn=audio_chunk_worker_init_fn,
        ... )

    """

    def __init__(
        self, recordings: RecordingSet, chunk_size: Seconds, chunk_shift: Seconds
    ) -> None:
        self.recordings = list(recordings)
        self.chunk_size = chunk_size
        self.chunk_shift = chunk_shift
        self.start = 0
        self.end = len(self.recordings)

        self.validate()

    def validate(self) -> None:
        for r in self.recordings:
            assert (
                len(r.sources) == 1
            ), f"We currently don't support multi-source audio in this dataset (got {len(r.sources)} sources in recording {r.id})."
            assert (
                r.sources[0].type == "file"
            ), f"We currently only support 'file' AudioSource type in this dataset (got: {r.sources[0].type} in recording {r.id})."
            assert (
                r.num_channels == 1
            ), f"We currently only support single-channel audio in this dataset (got {r.num_channels} channels in recording {r.id})."

    def __iter__(self):
        import soundfile as sf

        for r in self.recordings[self.start : self.end]:
            chunk_size = compute_num_samples(self.chunk_size, r.sampling_rate)
            chunk_overlap = compute_num_samples(
                self.chunk_size - self.chunk_shift, r.sampling_rate
            )

            begin_time = 0
            end_time = self.chunk_size
            with sf.SoundFile(r.sources[0].source, "rb") as stream:
                for chunk in stream.blocks(
                    chunk_size, overlap=chunk_overlap, dtype=np.float32
                ):
                    chunk = torch.as_tensor(chunk)
                    yield {
                        "recording_id": r.id,
                        "begin_time": torch.as_tensor(begin_time, dtype=torch.float32),
                        "end_time": torch.as_tensor(end_time, dtype=torch.float32),
                        "audio": chunk,
                    }
                    begin_time += self.chunk_shift
                    end_time = begin_time + self.chunk_size


def audio_chunk_collate(batch: List[Dict]):
    from torch.utils.data import default_collate

    audios = [d.pop("audio") for d in batch]
    out = default_collate(batch)

    maxlen = max(a.shape[0] for a in audios)
    audio = torch.zeros((len(audios), maxlen))
    for i, a in enumerate(audios):
        audio[i, : a.shape[0]] = torch.as_tensor(a)
    out["audio"] = audio

    return out


def audio_chunk_worker_init_fn(worker_id):
    worker_info = torch.utils.data.get_worker_info()
    dataset = worker_info.dataset  # the dataset copy in this worker process
    overall_start = dataset.start
    overall_end = dataset.end
    # configure the dataset to only process the split workload
    per_worker = int(
        math.ceil((overall_end - overall_start) / float(worker_info.num_workers))
    )
    worker_id = worker_info.id
    dataset.start = overall_start + worker_id * per_worker
    dataset.end = min(dataset.start + per_worker, overall_end)



"""
About the Bengali.AI Speech corpus

The competition dataset comprises about 1200 hours of recordings of Bengali speech.
Your goal is to transcribe recordings of speech that is out-of-distribution with respect to the training set.

Note that this is a Code Competition, in which the actual test set is hidden.
In this public version, we give some sample data in the correct format to help you author your solutions.
The full test set contains about 20 hours of speech in almost 8000 MP3 audio files.
All of the files in the test set are encoded at a sample rate of 32k, a bit rate of 48k, in one channel.

It is covered in more detail at https://arxiv.org/abs/2305.09688

Please download manually by
kaggle competitions download -c bengaliai-speech
"""

import logging
import os
from collections import defaultdict
from concurrent.futures.process import ProcessPoolExecutor
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import (
    get_ffmpeg_torchaudio_info_enabled,
    set_ffmpeg_torchaudio_info_enabled,
)
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.recipes.utils import manifests_exist
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike

BENGALIAI_SPEECH = ("train", "valid", "test")


@contextmanager
def disable_ffmpeg_torchaudio_info() -> None:
    enabled = get_ffmpeg_torchaudio_info_enabled()
    set_ffmpeg_torchaudio_info_enabled(False)
    try:
        yield
    finally:
        set_ffmpeg_torchaudio_info_enabled(enabled)


def _parse_utterance(
    corpus_dir: Pathlike,
    audio_path: Pathlike,
    audio_id: str,
    text: Optional[str] = "",
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    audio_path = audio_path.resolve()

    if not audio_path.is_file():
        logging.warning(f"No such file: {audio_path}")
        return None

    recording = Recording.from_file(
        path=audio_path,
        recording_id=audio_id,
    )
    segment = SupervisionSegment(
        id=audio_id,
        recording_id=audio_id,
        text=text,
        start=0.0,
        duration=recording.duration,
        channel=0,
        language="Bengali",
    )

    return recording, segment


def _prepare_subset(
    subset: str,
    corpus_dir: Pathlike,
    audio_info: Optional[dict] = None,
    num_jobs: int = 1,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param subset: str, the name of the subset.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    if subset == "test":
        part_path = corpus_dir / "test_mp3s"
    else:
        part_path = corpus_dir / "train_mp3s"

    audio_paths = list(part_path.rglob("*.mp3"))

    with disable_ffmpeg_torchaudio_info():
        with ProcessPoolExecutor(num_jobs) as ex:
            futures = []
            recordings = []
            supervisions = []
            for audio_path in tqdm(audio_paths, desc="Distributing tasks"):
                audio_id = os.path.split(str(audio_path))[1].replace(".mp3", "")
                if audio_info is not None:
                    if audio_id not in audio_info.keys():
                        continue
                    text = audio_info[audio_id]
                else:
                    text = None
                futures.append(
                    ex.submit(_parse_utterance, corpus_dir, audio_path, audio_id, text)
                )

            for future in tqdm(futures, desc="Processing"):
                result = future.result()
                if result is None:
                    continue
                recording, segment = result
                recordings.append(recording)
                supervisions.append(segment)

            recording_set = RecordingSet.from_recordings(recordings)
            supervision_set = SupervisionSet.from_segments(supervisions)

    return recording_set, supervision_set


def prepare_bengaliai_speech(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the Bengali.AI Speech dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing Bengali.AI Speech...")

    subsets = BENGALIAI_SPEECH

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    with open(corpus_dir / "train.csv") as f:
        audio_info = f.read().splitlines()

    train_info = {}
    valid_info = {}
    for line in audio_info[1:]:
        if ",train" in line:
            line = line.replace(",train", "").split(",", 1)
            train_info[line[0]] = line[1]
        elif ",valid" in line:
            line = line.replace(",valid", "").split(",", 1)
            valid_info[line[0]] = line[1]

    for part in tqdm(subsets, desc="Dataset parts"):
        logging.info(f"Processing Bengali.AI Speech subset: {part}")
        if manifests_exist(
            part=part,
            output_dir=output_dir,
            prefix="bengaliai_speech",
            suffix="jsonl.gz",
        ):
            logging.info(
                f"Bengali.AI Speech subset: {part} already prepared - skipping."
            )
            continue

        recording_set, supervision_set = _prepare_subset(
            subset=part,
            corpus_dir=corpus_dir,
            audio_info=train_info
            if part == "train"
            else valid_info
            if part == "valid"
            else None,
            num_jobs=num_jobs,
        )

        # Fix manifests
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"bengaliai_speech_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"bengaliai_speech_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests
